"""SDV Exceptions."""


class NotFittedError(Exception):
    """Error to raise when sample is called and the model is not fitted."""


class ConstraintsNotMetError(ValueError):
    """Exception raised when the given data is not valid for the constraints."""

    def __init__(self, message=''):
        self.message = message
        super().__init__(self.message)



import os
import zipfile

import gdown
from dotenv import load_dotenv

load_dotenv()

PROJECT_PATH = os.getenv("PROJECT_PATH")


def download_and_extract(url, filename):
    # Path to the directory where the file will be extracted
    extract_dir = os.path.join(PROJECT_PATH)

    # Create the download directory if it doesn't exist
    os.makedirs(extract_dir, exist_ok=True)

    file_path = os.path.join(extract_dir, filename)
    gdown.download(url, file_path, quiet=False)

    # Extract the file
    with zipfile.ZipFile(file_path, "r") as zip_ref:
        zip_ref.extractall(extract_dir)

    # Clean up the downloaded zip file
    os.remove(file_path)


# URL of the file to download
results_url = "https://drive.google.com/uc?id=1FprsHPYWDrE2AJKscRIVET_fvh0LoJX1"

download_and_extract(results_url, "results.zip")

print("Results downloaded and extracted successfully!")



"""Base Constraint class."""

import copy
import importlib
import inspect
import logging
import warnings

import pandas as pd
from copulas.multivariate.gaussian import GaussianMultivariate
from copulas.univariate import GaussianUnivariate
from rdt import HyperTransformer

from rctgan.constraints.errors import MissingConstraintColumnError

LOGGER = logging.getLogger(__name__)


def _get_qualified_name(obj):
    """Return the Fully Qualified Name from an instance or class."""
    module = obj.__module__
    if hasattr(obj, '__name__'):
        obj_name = obj.__name__
    else:
        obj_name = obj.__class__.__name__

    return module + '.' + obj_name


def _module_contains_callable_name(obj):
    """Return if module contains the name of the callable object."""
    if hasattr(obj, '__name__'):
        obj_name = obj.__name__
    else:
        obj_name = obj.__class__.__name__
    return obj_name in importlib.import_module(obj.__module__).__dict__


def get_subclasses(cls):
    """Recursively find subclasses for the current class object."""
    subclasses = dict()
    for subclass in cls.__subclasses__():
        subclasses[subclass.__name__] = subclass
        subclasses.update(get_subclasses(subclass))

    return subclasses


def import_object(obj):
    """Import an object from its qualified name."""
    if isinstance(obj, str):
        package, name = obj.rsplit('.', 1)
        return getattr(importlib.import_module(package), name)

    return obj


class ConstraintMeta(type):
    """Metaclass for Constraints.

    This metaclass replaces the ``__init__`` method with a new function
    that stores the arguments passed to the __init__ method in a dict
    as the attribute ``__kwargs__``.

    This allows us to later on dump the class definition as a dict.
    """

    def __init__(self, name, bases, attr):
        super().__init__(name, bases, attr)

        old__init__ = self.__init__
        signature = inspect.signature(old__init__)
        arg_names = list(signature.parameters.keys())[1:]

        def __init__(self, *args, **kwargs):
            class_name = self.__class__.__name__
            if name == class_name:
                self.__kwargs__ = copy.deepcopy(kwargs)
                self.__kwargs__.update(dict(zip(arg_names, args)))

            old__init__(self, *args, **kwargs)

        __init__.__doc__ = old__init__.__doc__
        __init__.__signature__ = signature
        self.__init__ = __init__


class Constraint(metaclass=ConstraintMeta):
    """Constraint base class.

    This class is not intended to be used directly and should rather be
    subclassed to create different types of constraints.

    If ``handling_strategy`` is passed with the value ``transform``
    or ``reject_sampling``, the ``filter_valid`` or ``transform`` and
    ``reverse_transform`` methods will be replaced respectively by a simple
    identity function.

    Attributes:
        constraint_columns (tuple[str]):
            The names of the columns used by this constraint.
        rebuild_columns (typle[str]):
            The names of the columns that this constraint will rebuild during
            ``reverse_transform``.
    Args:
        handling_strategy (str):
            How this Constraint should be handled, which can be ``transform``,
            ``reject_sampling`` or ``all``.
        fit_columns_model (bool):
            If False, reject sampling will be used to handle conditional sampling.
            Otherwise, a model will be trained and used to sample other columns
            based on the conditioned column.
    """

    constraint_columns = ()
    rebuild_columns = ()
    _hyper_transformer = None
    _columns_model = None

    def _identity(self, table_data):
        return table_data

    def __init__(self, handling_strategy, fit_columns_model=False):
        self.fit_columns_model = fit_columns_model
        if handling_strategy == 'transform':
            self.filter_valid = self._identity
        elif handling_strategy == 'reject_sampling':
            self.rebuild_columns = ()
            self.transform = self._identity
            self.reverse_transform = self._identity
        elif handling_strategy != 'all':
            raise ValueError('Unknown handling strategy: {}'.format(handling_strategy))

    def _fit(self, table_data):
        del table_data

    def fit(self, table_data):
        """Fit ``Constraint`` class to data.

        If ``fit_columns_model`` is True, then this method will fit
        a ``GaussianCopula`` model to the relevant columns in ``table_data``.
        Subclasses can overwrite this method, or overwrite the ``_fit`` method
        if they will not be needing the model to handle conditional sampling.

        Args:
            table_data (pandas.DataFrame):
                Table data.
        """
        self._fit(table_data)

        if self.fit_columns_model and len(self.constraint_columns) > 1:
            data_to_model = table_data[list(self.constraint_columns)]
            self._hyper_transformer = HyperTransformer(default_data_type_transformers={
                'categorical': 'OneHotEncodingTransformer',
            })
            transformed_data = self._hyper_transformer.fit_transform(data_to_model)
            self._columns_model = GaussianMultivariate(
                distribution=GaussianUnivariate
            )
            self._columns_model.fit(transformed_data)

    def _transform(self, table_data):
        return table_data

    def _reject_sample(self, num_rows, conditions):
        sampled = self._columns_model.sample(
            num_rows=num_rows,
            conditions=conditions
        )
        sampled = self._hyper_transformer.reverse_transform(sampled)
        valid_rows = sampled[self.is_valid(sampled)]
        counter = 0
        total_sampled = num_rows

        while len(valid_rows) < num_rows:
            num_valid = len(valid_rows)
            if counter >= 100:
                if len(valid_rows) == 0:
                    error = 'Could not get enough valid rows within 100 trials.'
                    raise ValueError(error)
                else:
                    multiplier = num_rows // num_valid
                    num_rows_missing = num_rows % num_valid
                    remainder_rows = valid_rows.iloc[0:num_rows_missing, :]
                    valid_rows = pd.concat([valid_rows] * multiplier + [remainder_rows],
                                           ignore_index=True)
                    break

            remaining = num_rows - num_valid
            valid_probability = (num_valid + 1) / (total_sampled + 1)
            max_rows = num_rows * 10
            num_to_sample = min(int(remaining / valid_probability), max_rows)
            total_sampled += num_to_sample
            new_sampled = self._columns_model.sample(
                num_rows=num_to_sample,
                conditions=conditions
            )
            new_sampled = self._hyper_transformer.reverse_transform(new_sampled)
            new_valid_rows = new_sampled[self.is_valid(new_sampled)]
            valid_rows = pd.concat([valid_rows, new_valid_rows], ignore_index=True)
            counter += 1

        return valid_rows.iloc[0:num_rows, :]

    def _sample_constraint_columns(self, table_data):
        condition_columns = [c for c in self.constraint_columns if c in table_data.columns]
        grouped_conditions = table_data[condition_columns].groupby(condition_columns)
        all_sampled_rows = list()
        for group, df in grouped_conditions:
            if not isinstance(group, tuple):
                group = [group]

            transformed_condition = self._hyper_transformer.transform(df).iloc[0].to_dict()
            sampled_rows = self._reject_sample(
                num_rows=df.shape[0],
                conditions=transformed_condition
            )
            all_sampled_rows.append(sampled_rows)

        sampled_data = pd.concat(all_sampled_rows, ignore_index=True)
        return sampled_data

    def _validate_constraint_columns(self, table_data):
        """Validate the columns in ``table_data``.

        If ``fit_columns_model`` is False and any columns in ``constraint_columns``
        are not present in ``table_data``, this method will raise a
        ``MissingConstraintColumnError``. Otherwise it will return the ``table_data``
        unchanged. If ``fit_columns_model`` is True, then this method will sample
        any missing ``constraint_columns`` from its model conditioned on the
        ``constraint_columns`` that ``table_data`` does contain. If ``table_data``
        doesn't contain any of the ``constraint_columns`` then a
        ``MissingConstraintColumnError`` will be raised.

        Args:
            table_data (pandas.DataFrame):
                Table data.
        """
        missing_columns = [col for col in self.constraint_columns if col not in table_data.columns]
        if missing_columns:
            if not self._columns_model:
                warning_message = (
                    'When `fit_columns_model` is False and we are conditioning on a subset '
                    'of the constraint columns, conditional sampling uses reject sampling '
                    'which can be slow. Changing `fit_columns_model` to True can improve '
                    'the performance.'
                )
                warnings.warn(warning_message, UserWarning)

            all_columns_missing = len(missing_columns) == len(self.constraint_columns)
            if self._columns_model is None or all_columns_missing:
                raise MissingConstraintColumnError()

            else:
                sampled_data = self._sample_constraint_columns(table_data)
                other_columns = [c for c in table_data.columns if c not in self.constraint_columns]
                sampled_data[other_columns] = table_data[other_columns]
                return sampled_data

        return table_data

    def transform(self, table_data):
        """Perform necessary transformations needed by constraint.

        Subclasses can optionally overwrite this method. If the transformation
        requires certain columns to be present in ``table_data``, then the subclass
        should overwrite the ``_transform`` method instead. This method raises a
        ``MissingConstraintColumnError`` if the ``table_data`` is missing any columns
        needed to do the transformation. If columns are present, this method will call
        the ``_transform`` method.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Input data unmodified.
        """
        table_data = self._validate_constraint_columns(table_data)
        return self._transform(table_data)

    def fit_transform(self, table_data):
        """Fit this Constraint to the data and then transform it.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        self.fit(table_data)
        return self.transform(table_data)

    def reverse_transform(self, table_data):
        """Identity method for completion. To be optionally overwritten by subclasses.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Input data unmodified.
        """
        return table_data

    def is_valid(self, table_data):
        """Say whether the given table rows are valid.

        This is a dummy version of the method that returns a series of ``True``
        values to avoid dropping any rows. This should be overwritten by all
        the subclasses that have a way to decide which rows are valid and which
        are not.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.Series:
                Series of ``True`` values
        """
        return pd.Series(True, index=table_data.index)

    def filter_valid(self, table_data):
        """Get only the rows that are valid.

        The filtering is done by calling the method ``is_valid``, which should
        be overwritten by subclasses, while this method should stay untouched.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Input data unmodified.
        """
        valid = self.is_valid(table_data)
        invalid = sum(~valid)
        if invalid:
            LOGGER.debug('%s: %s invalid rows out of %s.',
                         self.__class__.__name__, sum(~valid), len(valid))

        if isinstance(valid, pd.Series):
            return table_data[valid.values]

        return table_data[valid]

    @classmethod
    def from_dict(cls, constraint_dict):
        """Build a Constraint object from a dict.

        Args:
            constraint_dict (dict):
                Dict containing the keyword ``constraint`` alongside
                any additional arguments needed to create the instance.

        Returns:
            Constraint:
                New constraint instance.
        """
        constraint_dict = constraint_dict.copy()
        constraint_class = constraint_dict.pop('constraint')
        subclasses = get_subclasses(cls)
        if isinstance(constraint_class, str):
            if '.' in constraint_class:
                constraint_class = import_object(constraint_class)
            else:
                constraint_class = subclasses[constraint_class]

        return constraint_class(**constraint_dict)

    def to_dict(self):
        """Return a dict representation of this Constraint.

        The dictionary will contain the Qualified Name of the constraint
        class in the key ``constraint``, as well as any other arguments
        that were passed to the constructor when the instance was created.

        Returns:
            dict:
                Dict representation of this Constraint.
        """
        constraint_dict = {
            'constraint': _get_qualified_name(self.__class__),
        }

        for key, obj in copy.deepcopy(self.__kwargs__).items():
            if callable(obj) and _module_contains_callable_name(obj):
                constraint_dict[key] = _get_qualified_name(obj)
            else:
                constraint_dict[key] = obj

        return constraint_dict



import os
import sys
import glob
import logging
import argparse
from pathlib import Path


from realtabformer import REaLTabFormer
from syntherela.metadata import Metadata
from syntherela.data import load_tables, save_tables, remove_sdv_columns

MODEL_NAME = "REALTABFORMER"
BATCH_SIZE_PARENT = 1024
BATCH_SIZE_CHILD = 1

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="airbnb-simplified_subsampled")
args.add_argument("--real-data-path", type=str, default="data/original")
args.add_argument("--synthetic-data-path", type=str, default="data/synthetic")
args.add_argument("--full-sensitivity", type=bool, default=True)
args.add_argument("--retrain", type=bool, default=True)
args.add_argument("--run-id", type=str, default="1")
args = args.parse_args()

dataset_name = args.dataset_name
full_sensitivity = args.full_sensitivity
retrain = args.retrain
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
run_id = args.run_id

logger = logging.getLogger(f"{MODEL_NAME}_logger")

logger.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.info(f"START LOGGING Dataset: {dataset_name} RUN {run_id}...")

os.environ["WANDB_PROJECT"] = f"{MODEL_NAME}_{dataset_name}_run_{run_id}"
os.environ["WANDB__SERVICE_WAIT"] = "300"

report_to = None

os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "expandable_segments:True"


logger.info(f"START {MODEL_NAME}...")


logger.debug("Loading real data...")
metadata = Metadata().load_from_json(
    Path(real_data_path) / f"{dataset_name}/metadata.json"
)
real_data = load_tables(Path(real_data_path) / f"{dataset_name}", metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata)
metadata.validate_data(real_data)
logger.debug("Real data loaded")

synthetic_data = {}

added_keys = {}

for table in real_data:
    if metadata.get_primary_key(table) is None:
        logger.debug(f"Table {table} has no primary key. Adding one...")
        real_data[table].reset_index(inplace=True)
        added_keys[table] = "index"
        logger.debug(f"Primary key added to table {table}")

renamed_keys = {}
join_on_name = "join_on_RTF"
for relationship in metadata.relationships:
    parent_key = relationship["parent_primary_key"]
    child_key = relationship["child_foreign_key"]

    if parent_key == child_key:
        continue

    # rename child key in child table to parent key
    real_data[relationship["parent_table_name"]].rename(
        columns={parent_key: join_on_name}, inplace=True
    )
    real_data[relationship["child_table_name"]].rename(
        columns={child_key: join_on_name}, inplace=True
    )
    renamed_keys[relationship["child_table_name"]] = child_key
    renamed_keys[relationship["parent_table_name"]] = parent_key
    logger.debug(f"Renamed {child_key} and {parent_key} to {join_on_name}")

# GENERATE SYNTHETIC DATA ---------------------------------

root_table_name = metadata.get_root_tables()[0]
parent_df = real_data[root_table_name]

join_on = (
    metadata.get_primary_key(root_table_name)
    if len(renamed_keys.keys()) == 0
    else join_on_name
)
parent_model_path = (
    f"models/{MODEL_NAME}/{dataset_name}/checkpoint_{root_table_name}_run_{run_id}"
)

if retrain or not (
    os.path.exists(parent_model_path) and len(os.listdir(parent_model_path)) > 0
):
    logger.debug("Starting parent table generation...")
    logger.debug("Fitting parent model...")
    parent_model = REaLTabFormer(
        model_type="tabular",
        batch_size=BATCH_SIZE_PARENT,
        checkpoints_dir=f"checkpoints/{MODEL_NAME}/{dataset_name}/checkpoint_{root_table_name}_run_{run_id}",
        random_state=int(run_id),
        report_to=report_to,
    )
    os.makedirs(parent_model_path, exist_ok=True)
    # fit and save parent model
    parent_model.fit(parent_df.drop(join_on, axis=1), full_sensitivity=full_sensitivity)
    parent_model.save(parent_model_path)
    logger.debug("Parent fitted and saved")
else:
    logger.debug("Skipping parent table generation and reading last trained model...")

# load trained parent model
directories = list(filter(os.path.isdir, glob.glob(f"{parent_model_path}/id*")))
directories.sort(key=lambda x: os.path.getmtime(x))
parent_model_path = directories[-1]
parent_model = REaLTabFormer(
    model_type="tabular",
    parent_realtabformer_path=parent_model_path,
    batch_size=BATCH_SIZE_PARENT,
    checkpoints_dir=f"checkpoints/{MODEL_NAME}/{dataset_name}/checkpoint_{root_table_name}_run_{run_id}",
    random_state=int(run_id),
    report_to=report_to,
)

parent_model = parent_model.load_from_dir(parent_model_path)

for i in range(1, 4):
    # sample from parent model
    logger.debug(f"Sampling from parent model, table {root_table_name}...")
    parent_samples = parent_model.sample(real_data[root_table_name].shape[0])
    parent_samples.index.name = join_on
    parent_samples = parent_samples.reset_index()
    synthetic_data[root_table_name] = parent_samples
    logger.debug("Done.")

    # GENERATE CHILD TABLES
    for child_table_name in sorted(metadata.get_children(root_table_name)):
        logger.info(f"Starting child table generation for {child_table_name}...")
        child_model_path = f"models/{MODEL_NAME}/{dataset_name}/checkpoint_{child_table_name}_run_{run_id}"
        child_df = real_data[child_table_name]
        assert (join_on in parent_df.columns) and (join_on in child_df.columns)
        if retrain or not (
            os.path.exists(child_model_path) and len(os.listdir(child_model_path)) > 0
        ):
            child_model = REaLTabFormer(
                model_type="relational",
                parent_realtabformer_path=parent_model_path,
                output_max_length=None,
                train_size=0.95,
                batch_size=BATCH_SIZE_CHILD,
                checkpoints_dir=f"checkpoints/{MODEL_NAME}/{dataset_name}/checkpoint_{child_table_name}_run_{run_id}",
                random_state=int(run_id),
                report_to=report_to,
            )

            logger.debug(f"Fitting child model for table {child_table_name}...")
            # fit child model
            child_model.fit(df=child_df, in_df=parent_df, join_on=join_on)
            # save child model
            os.makedirs(child_model_path, exist_ok=True)
            child_model.save(child_model_path)
            logger.debug("Child fitted and saved")
        else:
            logger.debug(
                f"Skipping child table {child_table_name} generation and reading last trained model..."
            )

        directories = list(filter(os.path.isdir, glob.glob(f"{child_model_path}/id*")))
        directories.sort(key=lambda x: os.path.getmtime(x))
        child_model_path = directories[-1]
        child_model = REaLTabFormer(
            model_type="relational",
            parent_realtabformer_path=parent_model_path,
            output_max_length=None,
            train_size=0.95,
            batch_size=BATCH_SIZE_CHILD,
            checkpoints_dir=f"checkpoints/{MODEL_NAME}/{dataset_name}/checkpoint_{child_table_name}_run_{run_id}",
            random_state=int(run_id),
            report_to=report_to,
        )
        child_model = child_model.load_from_dir(child_model_path)

        logger.info("Starting sampling and saving synthetic data...")

        logger.debug(f"Sampling from child model, table {child_table_name}...")
        child_samples = child_model.sample(
            input_unique_ids=parent_samples[join_on],
            input_df=parent_samples.drop(join_on, axis=1),
            gen_batch=64,
        )

        child_samples.index.name = join_on
        child_samples = child_samples.reset_index()

        child_primary_key = (
            metadata.get_primary_key(child_table_name)
            if child_table_name not in added_keys
            else added_keys[child_table_name]
        )
        child_samples = child_samples.drop(columns=child_primary_key)
        child_samples.index.name = child_primary_key
        child_samples.reset_index(inplace=True)

        synthetic_data[child_table_name] = child_samples

        for table in synthetic_data:
            if table in added_keys:
                try:
                    synthetic_data[table].drop(columns=added_keys[table], inplace=True)
                except:  # noqa E722
                    pass
            if table in renamed_keys:
                synthetic_data[table].rename(
                    columns={join_on_name: renamed_keys[table]}, inplace=True
                )

        logger.info("Saving synthetic data...")
        save_data_path = (
            Path(synthetic_data_path)
            / dataset_name
            / MODEL_NAME
            / run_id
            / f"sample{i}"
        )
        save_tables(synthetic_data, save_data_path)
        logger.debug("Done.")

logger.info("COMPLETE GENERATION DONE.")



"""Base Class for tabular models."""

import logging
import pickle
import uuid
from warnings import warn

import numpy as np
import pandas as pd

from rctgan.errors import ConstraintsNotMetError
from rctgan.metadata import Table

LOGGER = logging.getLogger(__name__)
COND_IDX = str(uuid.uuid4())


class NonParametricError(Exception):
    """Exception to indicate that a model is not parametric."""


class BaseTabularModel:
    """Base class for all the tabular models.

    The ``BaseTabularModel`` class defines the common API that all the
    TabularModels need to implement, as well as common functionality.

    Args:
        field_names (list[str]):
            List of names of the fields that need to be modeled
            and included in the generated output data. Any additional
            fields found in the data will be ignored and will not be
            included in the generated output.
            If ``None``, all the fields found in the data are used.
        field_types (dict[str, dict]):
            Dictinary specifying the data types and subtypes
            of the fields that will be modeled. Field types and subtypes
            combinations must be compatible with the SDV Metadata Schema.
        field_transformers (dict[str, str]):
            Dictinary specifying which transformers to use for each field.
            Available transformers are:

                * ``integer``: Uses a ``NumericalTransformer`` of dtype ``int``.
                * ``float``: Uses a ``NumericalTransformer`` of dtype ``float``.
                * ``categorical``: Uses a ``CategoricalTransformer`` without gaussian noise.
                * ``categorical_fuzzy``: Uses a ``CategoricalTransformer`` adding gaussian noise.
                * ``one_hot_encoding``: Uses a ``OneHotEncodingTransformer``.
                * ``label_encoding``: Uses a ``LabelEncodingTransformer``.
                * ``boolean``: Uses a ``BooleanTransformer``.
                * ``datetime``: Uses a ``DatetimeTransformer``.

        anonymize_fields (dict[str, str]):
            Dict specifying which fields to anonymize and what faker
            category they belong to.
        primary_key (str):
            Name of the field which is the primary key of the table.
        constraints (list[Constraint, dict]):
            List of Constraint objects or dicts.
        table_metadata (dict or metadata.Table):
            Table metadata instance or dict representation.
            If given alongside any other metadata-related arguments, an
            exception will be raised.
            If not given at all, it will be built using the other
            arguments or learned from the data.
        rounding (int, str or None):
            Define rounding scheme for ``NumericalTransformer``. If set to an int, values
            will be rounded to that number of decimal places. If ``None``, values will not
            be rounded. If set to ``'auto'``, the transformer will round to the maximum number
            of decimal places detected in the fitted data. Defaults to ``'auto'``.
        min_value (int, str or None):
            Specify the minimum value the ``NumericalTransformer`` should use. If an integer
            is given, sampled data will be greater than or equal to it. If the string ``'auto'``
            is given, the minimum will be the minimum value seen in the fitted data. If ``None``
            is given, there won't be a minimum. Defaults to ``'auto'``.
        max_value (int, str or None):
            Specify the maximum value the ``NumericalTransformer`` should use. If an integer
            is given, sampled data will be less than or equal to it. If the string ``'auto'``
            is given, the maximum will be the maximum value seen in the fitted data. If ``None``
            is given, there won't be a maximum. Defaults to ``'auto'``.
    """

    _DTYPE_TRANSFORMERS = None

    _metadata = None

    def __init__(self, field_names=None, field_types=None, field_transformers=None,
                 anonymize_fields=None, primary_key=None, constraints=None, table_metadata=None,
                 rounding='auto', min_value='auto', max_value='auto'):
        if table_metadata is None:
            self._metadata = Table(
                field_names=field_names,
                primary_key=primary_key,
                field_types=field_types,
                field_transformers=field_transformers,
                anonymize_fields=anonymize_fields,
                constraints=constraints,
                dtype_transformers=self._DTYPE_TRANSFORMERS,
                rounding=rounding,
                min_value=min_value,
                max_value=max_value
            )
            self._metadata_fitted = False
        else:
            for arg in (field_names, primary_key, field_types, anonymize_fields, constraints):
                if arg:
                    raise ValueError(
                        'If table_metadata is given {} must be None'.format(arg.__name__))

            if isinstance(table_metadata, dict):
                table_metadata = Table.from_dict(table_metadata)

            table_metadata._dtype_transformers.update(self._DTYPE_TRANSFORMERS)

            self._metadata = table_metadata
            self._metadata_fitted = table_metadata.fitted

    def fit(self, data):
        """Fit this model to the data.

        If the table metadata has not been given, learn it from the data.

        Args:
            data (pandas.DataFrame or str):
                Data to fit the model to. It can be passed as a
                ``pandas.DataFrame`` or as an ``str``.
                If an ``str`` is passed, it is assumed to be
                the path to a CSV file which can be loaded using
                ``pandas.read_csv``.
        """
        if isinstance(data, pd.DataFrame):
            data = data.reset_index(drop=True)

        LOGGER.debug('Fitting %s to table %s; shape: %s', self.__class__.__name__,
                     self._metadata.name, data.shape)
        if not self._metadata_fitted:
            self._metadata.fit(data)

        self._num_rows = len(data)

        LOGGER.debug('Transforming table %s; shape: %s', self._metadata.name, data.shape)
        transformed = self._metadata.transform(data)

        if self._metadata.get_dtypes(ids=False):
            LOGGER.debug(
                'Fitting %s model to table %s', self.__class__.__name__, self._metadata.name)
            self._fit(transformed)

    def get_metadata(self):
        """Get metadata about the table.

        This will return an ``sdv.metadata.Table`` object containing
        the information about the data that this model has learned.

        This Table metadata will contain some common information,
        such as field names and data types, as well as additional
        information that each Sub-class might add, such as the
        observed data field distributions and their parameters.

        Returns:
            sdv.metadata.Table:
                Table metadata.
        """
        return self._metadata

    @staticmethod
    def _filter_conditions(sampled, conditions, float_rtol):
        """Filter the sampled rows that match the conditions.

        If condition columns are float values, consider a match anything that
        is closer than the given ``float_rtol`` and then make the value exact.

        Args:
            sampled (pandas.DataFrame):
                The sampled rows, reverse transformed.
            conditions (dict):
                The dictionary of conditioning values.
            float_rtol (float):
                Maximum tolerance when considering a float match.

        Returns:
            pandas.DataFrame:
                Rows from the sampled data that match the conditions.
        """
        for column, value in conditions.items():
            column_values = sampled[column]
            if column_values.dtype.kind == 'f':
                distance = value * float_rtol
                sampled = sampled[np.abs(column_values - value) < distance]
                sampled[column] = value
            else:
                sampled = sampled[column_values == value]

        return sampled

    def _sample_rows(self, num_rows, conditions=None, transformed_conditions=None,
                     float_rtol=0.1, previous_rows=None):
        """Sample rows with the given conditions.

        Input conditions is taken both in the raw input format, which will be used
        for filtering during the reject-sampling loop, and already transformed
        to the model format, which will be passed down to the model if it supports
        conditional sampling natively.

        If condition columns are float values, consider a match anything that
        is closer than the given ``float_rtol`` and then make the value exact.

        If the model does not have any data columns, the result of this call
        is a dataframe of the requested length with no columns in it.

        Args:
            num_rows (int):
                Number of rows to sample.
            conditions (dict):
                The dictionary of conditioning values in the original format.
            transformed_conditions (dict):
                The dictionary of conditioning values transformed to the model format.
            float_rtol (float):
                Maximum tolerance when considering a float match.
            previous_rows (pandas.DataFrame):
                Valid rows sampled in the previous iterations.

        Returns:
            tuple:
                * pandas.DataFrame:
                    Rows from the sampled data that match the conditions.
                * int:
                    Number of rows that are considered valid.
        """
        if self._metadata.get_dtypes(ids=False):
            if conditions is None:
                sampled = self._sample(num_rows)
            else:
                try:
                    sampled = self._sample(num_rows, transformed_conditions)
                except NotImplementedError:
                    sampled = self._sample(num_rows)

            sampled = self._metadata.reverse_transform(sampled)

            if previous_rows is not None:
                sampled = previous_rows.append(sampled, ignore_index=True)

            sampled = self._metadata.filter_valid(sampled)

            if conditions is not None:
                sampled = self._filter_conditions(sampled, conditions, float_rtol)

            num_valid = len(sampled)

            return sampled, num_valid

        else:
            sampled = pd.DataFrame(index=range(num_rows))
            sampled = self._metadata.reverse_transform(sampled)
            return sampled, num_rows

    def _sample_batch(self, num_rows=None, max_retries=100, max_rows_multiplier=10,
                      conditions=None, transformed_conditions=None, float_rtol=0.01):
        """Sample a batch of rows with the given conditions.

        This will enter a reject-sampling loop in which rows will be sampled until
        all of them are valid and match the requested conditions. If `max_retries`
        is exceeded, it will return as many rows as it has sampled, which may be less
        than the target number of rows.

        Input conditions is taken both in the raw input format, which will be used
        for filtering during the reject-sampling loop, and already transformed
        to the model format, which will be passed down to the model if it supports
        conditional sampling natively.

        If condition columns are float values, consider a match anything that is
        relatively closer than the given ``float_rtol`` and then make the value exact.

        If the model does not have any data columns, the result of this call
        is a dataframe of the requested length with no columns in it.

        Args:
            num_rows (int):
                Number of rows to sample. If not given the model
                will generate as many rows as there were in the
                data passed to the ``fit`` method.
            max_retries (int):
                Number of times to retry sampling discarded rows.
                Defaults to 100.
            max_rows_multiplier (int):
                Multiplier to use when computing the maximum number of rows
                that can be sampled during the reject-sampling loop.
                The maximum number of rows that are sampled at each iteration
                will be equal to this number multiplied by the requested num_rows.
                Defaults to 10.
            conditions (dict):
                The dictionary of conditioning values in the original input format.
            transformed_conditions (dict):
                The dictionary of conditioning values transformed to the model format.
            float_rtol (float):
                Maximum tolerance when considering a float match.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        sampled, num_valid = self._sample_rows(
            num_rows, conditions, transformed_conditions, float_rtol)

        counter = 0
        total_sampled = num_rows
        while num_valid < num_rows:
            if counter >= max_retries:
                break

            remaining = num_rows - num_valid
            valid_probability = (num_valid + 1) / (total_sampled + 1)
            max_rows = num_rows * max_rows_multiplier
            num_to_sample = min(int(remaining / valid_probability), max_rows)
            total_sampled += num_to_sample

            LOGGER.info('%s valid rows remaining. Resampling %s rows', remaining, num_to_sample)
            sampled, num_valid = self._sample_rows(
                num_to_sample, conditions, transformed_conditions, float_rtol, sampled
            )

            counter += 1

        return sampled.head(min(len(sampled), num_rows))

    def _make_conditions_df(self, conditions, num_rows):
        """Transform `conditions` into a dataframe.

        Args:
            conditions (pd.DataFrame, dict or pd.Series):
                If this is a dictionary/Series which maps column names to the column
                value, then this method generates `num_rows` samples, all of
                which are conditioned on the given variables. If this is a DataFrame,
                then it generates an output DataFrame such that each row in the output
                is sampled conditional on the corresponding row in the input.
            num_rows (int):
                Number of rows to sample. If a conditions dataframe is given, this must
                either be ``None`` or match the length of the ``conditions`` dataframe.

        Returns:
            pandas.DataFrame:
                `conditions` as a dataframe.
        """
        n_rows = num_rows if num_rows else self._num_rows
        if isinstance(conditions, pd.Series):
            conditions = pd.DataFrame([conditions] * n_rows)

        elif isinstance(conditions, dict):
            conditions = pd.DataFrame(conditions, index=range(n_rows))

        elif not isinstance(conditions, pd.DataFrame):
            raise TypeError('`conditions` must be a dataframe, a dictionary or a pandas series.')

        elif num_rows is not None and len(conditions) != num_rows:
            raise ValueError(
                'If `conditions` is a `DataFrame`, `num_rows` must be `None` or match its lenght.')

        return conditions.copy()

    def _conditionally_sample_rows(self, dataframe, max_retries, max_rows_multiplier,
                                   condition, transformed_condition, float_rtol,
                                   graceful_reject_sampling):
        num_rows = len(dataframe)
        sampled_rows = self._sample_batch(
            num_rows,
            max_retries,
            max_rows_multiplier,
            condition,
            transformed_condition,
            float_rtol
        )
        num_sampled_rows = len(sampled_rows)

        if num_sampled_rows < num_rows:
            # Didn't get enough rows.
            if len(sampled_rows) == 0:
                error = 'No valid rows could be generated with the given conditions.'
                raise ValueError(error)

            elif not graceful_reject_sampling:
                error = f'Could not get enough valid rows within {max_retries} trials.'
                raise ValueError(error)

            else:
                warn(f'Only {len(sampled_rows)} rows could '
                     f'be sampled within {max_retries} trials.')

        if len(sampled_rows) > 0:
            sampled_rows[COND_IDX] = dataframe[COND_IDX].values[:len(sampled_rows)]

        return sampled_rows

    def sample(self, num_rows=None, max_retries=100, max_rows_multiplier=10,
               conditions=None, float_rtol=0.01, graceful_reject_sampling=False):
        """Sample rows from this table.

        Args:
            num_rows (int):
                Number of rows to sample. If not given the model
                will generate as many rows as there were in the
                data passed to the ``fit`` method.
            max_retries (int):
                Number of times to retry sampling discarded rows.
                Defaults to 100.
            max_rows_multiplier (int):
                Multiplier to use when computing the maximum number of rows
                that can be sampled during the reject-sampling loop.
                The maximum number of rows that are sampled at each iteration
                will be equal to this number multiplied by the requested num_rows.
                Defaults to 10.
            conditions (pd.DataFrame, dict or pd.Series):
                If this is a dictionary/Series which maps column names to the column
                value, then this method generates `num_rows` samples, all of
                which are conditioned on the given variables. If this is a DataFrame,
                then it generates an output DataFrame such that each row in the output
                is sampled conditional on the corresponding row in the input.
            float_rtol (float):
                Maximum tolerance when considering a float match. This is the maximum
                relative distance at which a float value will be considered a match
                when performing reject-sampling based conditioning. Defaults to 0.01.
            graceful_reject_sampling (bool):
                If `False` raises a `ValueError` if not enough valid rows could be sampled
                within `max_retries` trials. If `True` prints a warning and returns
                as many rows as it was able to sample within `max_retries`.
                Defaults to False.

        Returns:
            pandas.DataFrame:
                Sampled data.

        Raises:
            ConstraintsNotMetError:
                If the conditions are not valid for the given constraints.
            ValueError:
                If any of the following happens:
                    * any of the conditions' columns are not valid.
                    * `graceful_reject_sampling` is `False` and not enough valid rows could be
                      sampled within `max_retries` trials.
                    * no rows could be generated.
        """
        if conditions is None:
            num_rows = num_rows or self._num_rows
            return self._sample_batch(num_rows, max_retries, max_rows_multiplier)

        # convert conditions to dataframe
        conditions = self._make_conditions_df(conditions, num_rows)

        # validate columns
        for column in conditions.columns:
            if column not in self._metadata.get_fields():
                raise ValueError(f'Invalid column name `{column}`')

        try:
            transformed_conditions = self._metadata.transform(conditions, on_missing_column='drop')
        except ConstraintsNotMetError as cnme:
            cnme.message = 'Passed conditions are not valid for the given constraints'
            raise

        condition_columns = list(conditions.columns)
        transformed_columns = list(transformed_conditions.columns)
        conditions.index.name = COND_IDX
        conditions.reset_index(inplace=True)
        transformed_conditions.index.name = COND_IDX
        transformed_conditions.reset_index(inplace=True)
        grouped_conditions = conditions.groupby(condition_columns)

        # sample
        all_sampled_rows = list()

        for group, dataframe in grouped_conditions:
            if not isinstance(group, tuple):
                group = [group]

            condition_indices = dataframe[COND_IDX]
            condition = dict(zip(condition_columns, group))
            if len(transformed_columns) == 0:
                sampled_rows = self._conditionally_sample_rows(
                    dataframe,
                    max_retries,
                    max_rows_multiplier,
                    condition,
                    None,
                    float_rtol,
                    graceful_reject_sampling
                )
                all_sampled_rows.append(sampled_rows)
            else:
                transformed_conditions_in_group = transformed_conditions.loc[condition_indices]
                transformed_groups = transformed_conditions_in_group.groupby(transformed_columns)
                for transformed_group, transformed_dataframe in transformed_groups:
                    if not isinstance(transformed_group, tuple):
                        transformed_group = [transformed_group]

                    transformed_condition = dict(zip(transformed_columns, transformed_group))
                    sampled_rows = self._conditionally_sample_rows(
                        transformed_dataframe,
                        max_retries,
                        max_rows_multiplier,
                        condition,
                        transformed_condition,
                        float_rtol,
                        graceful_reject_sampling
                    )
                    all_sampled_rows.append(sampled_rows)

        all_sampled_rows = pd.concat(all_sampled_rows)
        all_sampled_rows = all_sampled_rows.set_index(COND_IDX)
        all_sampled_rows.index.name = conditions.index.name
        all_sampled_rows = all_sampled_rows.sort_index()
        all_sampled_rows = self._metadata.make_ids_unique(all_sampled_rows)

        return all_sampled_rows

    def _get_parameters(self):
        raise NonParametricError()

    def get_parameters(self):
        """Get the parameters learned from the data.

        The result is a flat dict (single level) which contains
        all the necessary parameters to be able to reproduce
        this model.

        Subclasses which are not parametric, such as DeepLearning
        based models, raise a NonParametricError indicating that
        this method is not supported for their implementation.

        Returns:
            parameters (dict):
                flat dict (single level) which contains all the
                necessary parameters to be able to reproduce
                this model.

        Raises:
            NonParametricError:
                If the model is not parametric or cannot be described
                using a simple dictionary.
        """
        if self._metadata.get_dtypes(ids=False):
            parameters = self._get_parameters()
        else:
            parameters = {}

        parameters['num_rows'] = self._num_rows
        return parameters

    def _set_parameters(self, parameters):
        raise NonParametricError()

    def set_parameters(self, parameters):
        """Regenerate a previously learned model from its parameters.

        Subclasses which are not parametric, such as DeepLearning
        based models, raise a NonParametricError indicating that
        this method is not supported for their implementation.

        Args:
            dict:
                Model parameters.

        Raises:
            NonParametricError:
                If the model is not parametric or cannot be described
                using a simple dictionary.
        """
        num_rows = parameters.pop('num_rows')
        self._num_rows = 0 if pd.isnull(num_rows) else max(0, int(round(num_rows)))

        if self._metadata.get_dtypes(ids=False):
            self._set_parameters(parameters)

    def save(self, path):
        """Save this model instance to the given path using pickle.

        Args:
            path (str):
                Path where the SDV instance will be serialized.
        """
        with open(path, 'wb') as output:
            pickle.dump(self, output)

    @classmethod
    def load(cls, path):
        """Load a TabularModel instance from a given path.

        Args:
            path (str):
                Path from which to load the instance.

        Returns:
            TabularModel:
                The loaded tabular model.
        """
        with open(path, 'rb') as f:
            return pickle.load(f)



"""Synthesizers module."""

from rctgan.ctganpc.synthesizers.ctganpc import CTGANSynthesizer, PC_CTGANSynthesizer

__all__ = (
    'CTGANSynthesizer',
    'PC_CTGANSynthesizer'
)


def get_all_synthesizers():
    return {
        name: globals()[name]
        for name in __all__
    }



"""Metadata for a single table."""

import copy
import json
import logging

import numpy as np
import pandas as pd
import rdt
from faker import Faker

from rctgan.constraints.base import Constraint
from rctgan.constraints.errors import MissingConstraintColumnError
from rctgan.errors import ConstraintsNotMetError
from rctgan.metadata.errors import MetadataError, MetadataNotFittedError
from rctgan.metadata.utils import strings_from_regex

LOGGER = logging.getLogger(__name__)


class Table:
    """Table Metadata.

    The Metadata class provides a unified layer of abstraction over the metadata
    of a single Table, which includes all the necessary details to handle the
    table of this data, including the data types, the fields with pii information
    and the constraints that affect this data.

    Args:
        name (str):
            Name of this table. Optional.
        field_names (list[str]):
            List of names of the fields that need to be modeled
            and included in the generated output data. Any additional
            fields found in the data will be ignored and will not be
            included in the generated output.
            If ``None``, all the fields found in the data are used.
        field_types (dict[str, dict]):
            Dictinary specifying the data types and subtypes
            of the fields that will be modeled. Field types and subtypes
            combinations must be compatible with the SDV Metadata Schema.
        field_transformers (dict[str, str]):
            Dictinary specifying which transformers to use for each field.
            Available transformers are:

                * ``integer``: Uses a ``NumericalTransformer`` of dtype ``int``.
                * ``float``: Uses a ``NumericalTransformer`` of dtype ``float``.
                * ``categorical``: Uses a ``CategoricalTransformer`` without gaussian noise.
                * ``categorical_fuzzy``: Uses a ``CategoricalTransformer`` adding gaussian noise.
                * ``one_hot_encoding``: Uses a ``OneHotEncodingTransformer``.
                * ``label_encoding``: Uses a ``LabelEncodingTransformer``.
                * ``boolean``: Uses a ``BooleanTransformer``.
                * ``datetime``: Uses a ``DatetimeTransformer``.

        anonymize_fields (dict[str, str]):
            Dict specifying which fields to anonymize and what faker
            category they belong to.
        primary_key (str):
            Name of the field which is the primary key of the table.
        constraints (list[Constraint, dict]):
            List of Constraint objects or dicts.
        dtype_transformers (dict):
            Dictionary of transformer templates to be used for the
            different data types. The keys must be any of the `dtype.kind`
            values, `i`, `f`, `O`, `b` or `M`, and the values must be
            either RDT Transformer classes or RDT Transformer instances.
        model_kwargs (dict):
            Dictionary specifiying the kwargs that need to be used in
            each tabular model when working on this table. This dictionary
            contains as keys the name of the TabularModel class and as
            values a dictionary containing the keyword arguments to use.
            This argument exists mostly to ensure that the models are
            fitted using the same arguments when the same Table is used
            to fit different model instances on different slices of the
            same table.
        sequence_index (str):
            Name of the column that acts as the order index of each
            sequence. The sequence index column can be of any type that can
            be sorted, such as integer values or datetimes.
        entity_columns (list[str]):
            Names of the columns which identify different time series
            sequences. These will be used to group the data in separated
            training examples.
        context_columns (list[str]):
            The columns in the dataframe which are constant within each
            group/entity. These columns will be provided at sampling time
            (i.e. the samples will be conditioned on the context variables).
        rounding (int, str or None):
            Define rounding scheme for ``NumericalTransformer``. If set to an int, values
            will be rounded to that number of decimal places. If ``None``, values will not
            be rounded. If set to ``'auto'``, the transformer will round to the maximum number
            of decimal places detected in the fitted data. Defaults to ``'auto'``.
        min_value (int, str or None):
            Specify the minimum value the ``NumericalTransformer`` should use. If an integer
            is given, sampled data will be greater than or equal to it. If the string ``'auto'``
            is given, the minimum will be the minimum value seen in the fitted data. If ``None``
            is given, there won't be a minimum. Defaults to ``'auto'``.
        max_value (int, str or None):
            Specify the maximum value the ``NumericalTransformer`` should use. If an integer
            is given, sampled data will be less than or equal to it. If the string ``'auto'``
            is given, the maximum will be the maximum value seen in the fitted data. If ``None``
            is given, there won't be a maximum. Defaults to ``'auto'``.
    """

    _hyper_transformer = None
    _fakers = None
    _constraint_instances = None
    _fields_metadata = None
    fitted = False

    _ANONYMIZATION_MAPPINGS = dict()
    _TRANSFORMER_TEMPLATES = {
        'integer': rdt.transformers.NumericalTransformer(dtype=int),
        'float': rdt.transformers.NumericalTransformer(dtype=float),
        'categorical': rdt.transformers.CategoricalTransformer,
        'categorical_fuzzy': rdt.transformers.CategoricalTransformer(fuzzy=True),
        'one_hot_encoding': rdt.transformers.OneHotEncodingTransformer,
        'label_encoding': rdt.transformers.LabelEncodingTransformer,
        'boolean': rdt.transformers.BooleanTransformer,
        'datetime': rdt.transformers.DatetimeTransformer(strip_constant=True),
    }
    _DTYPE_TRANSFORMERS = {
        'i': 'integer',
        'f': 'float',
        'O': 'one_hot_encoding',
        'b': 'boolean',
        'M': 'datetime',
    }
    _DTYPES_TO_TYPES = {
        'i': {
            'type': 'numerical',
            'subtype': 'integer',
        },
        'f': {
            'type': 'numerical',
            'subtype': 'float',
        },
        'O': {
            'type': 'categorical',
        },
        'b': {
            'type': 'boolean',
        },
        'M': {
            'type': 'datetime',
        }
    }
    _TYPES_TO_DTYPES = {
        ('categorical', None): 'object',
        ('boolean', None): 'bool',
        ('numerical', None): 'float',
        ('numerical', 'float'): 'float',
        ('numerical', 'integer'): 'int',
        ('datetime', None): 'datetime64',
        ('id', None): 'int',
        ('id', 'integer'): 'int',
        ('id', 'string'): 'str'
    }

    @staticmethod
    def _get_faker(field_metadata):
        """Return the faker object with localisaton set if specified in field_metadata.

        Args:
            field_metadata (dict):
                Metadata for field to read localisation from if set in `pii_locales`.

        Returns:
            Faker object:
                The Faker object to anonymize the data in the field using its functions.
        """
        pii_locales = field_metadata.get('pii_locales', None)
        return Faker(locale=pii_locales)

    @staticmethod
    def _get_faker_method(faker, category):
        """Return the faker function to anonymize data.

        Args:
            faker (Faker object):
                The faker object created to get functions from.
            category (str or tuple):
                Fake category to use. If a tuple is passed, the first element is
                the category and the rest are additional arguments for the Faker.

        Returns:
            function:
                Faker function to generate new fake data instances.

        Raises:
            ValueError:
                A ``ValueError`` is raised if the faker category we want don't exist.
        """
        if isinstance(category, (tuple, list)):
            category, *args = category
        else:
            args = tuple()

        try:
            if args:
                def _faker():
                    return getattr(faker, category)(*args)

            else:
                def _faker():
                    return getattr(faker, category)()

            return _faker
        except AttributeError:
            raise ValueError('Category "{}" couldn\'t be found on faker'.format(category))

    @staticmethod
    def _get_fake_values(field_metadata, num_values):
        """Return the anonymized values from Faker.

        Args:
            field_metadata (dict):
                Metadata for field to read localisation from if set in `pii_locales`.
                And to read the faker category from `pii_category`.
            num_values (int):
                Number of values to create.

        Returns:
            generator:
                Generator containing the anonymized values.
        """
        faker = Table._get_faker(field_metadata)
        faker_method = Table._get_faker_method(faker, field_metadata['pii_category'])
        return (
            faker_method()
            for _ in range(num_values)
        )

    def _update_transformer_templates(self, rounding, min_value, max_value):
        default_numerical_transformer = self._TRANSFORMER_TEMPLATES['integer']
        if (rounding != default_numerical_transformer.rounding
                or min_value != default_numerical_transformer.min_value
                or max_value != default_numerical_transformer.max_value):
            custom_int = rdt.transformers.NumericalTransformer(
                dtype=int, rounding=rounding, min_value=min_value, max_value=max_value)
            custom_float = rdt.transformers.NumericalTransformer(
                dtype=float, rounding=rounding, min_value=min_value, max_value=max_value)
            self._transformer_templates.update({
                'integer': custom_int,
                'float': custom_float
            })

    @staticmethod
    def _prepare_constraints(constraints):
        constraints = constraints or []
        rebuild_columns = set()
        transform_constraints = []
        reject_sampling_constraints = []
        for constraint in constraints:
            if isinstance(constraint, type):
                constraint = constraint().to_dict()
            elif isinstance(constraint, Constraint):
                constraint = constraint.to_dict()

            constraint = Constraint.from_dict(constraint)

            if not constraint.rebuild_columns:
                reject_sampling_constraints.append(constraint)
            elif rebuild_columns & set(constraint.constraint_columns):
                intersecting_columns = rebuild_columns & set(constraint.constraint_columns)
                raise Exception('Multiple constraints will modify the same column(s): '
                                f'"{intersecting_columns}", which may lead to the constraint '
                                'being unenforceable. Please use "reject_sampling" as the '
                                '"handling_strategy" instead.')
            else:
                transform_constraints.append(constraint)
                rebuild_columns.update(constraint.rebuild_columns)

        return reject_sampling_constraints + transform_constraints

    def __init__(self, name=None, field_names=None, field_types=None, field_transformers=None,
                 anonymize_fields=None, primary_key=None, constraints=None,
                 dtype_transformers=None, model_kwargs=None, sequence_index=None,
                 entity_columns=None, context_columns=None, rounding=None, min_value=None,
                 max_value=None):
        self.name = name
        self._field_names = field_names
        self._field_types = field_types or {}
        self._field_transformers = field_transformers or {}
        self._anonymize_fields = anonymize_fields or {}
        self._model_kwargs = model_kwargs or {}

        self._primary_key = primary_key
        self._sequence_index = sequence_index
        self._entity_columns = entity_columns or []
        self._context_columns = context_columns or []
        self._constraints = self._prepare_constraints(constraints)
        self._dtype_transformers = self._DTYPE_TRANSFORMERS.copy()
        self._transformer_templates = self._TRANSFORMER_TEMPLATES.copy()
        self._update_transformer_templates(rounding, min_value, max_value)
        if dtype_transformers:
            self._dtype_transformers.update(dtype_transformers)

    def __repr__(self):
        return 'Table(name={}, field_names={})'.format(self.name, self._field_names)

    def get_model_kwargs(self, model_name):
        """Return the required model kwargs for the indicated model.

        Args:
            model_name (str):
                Qualified Name of the model for which model kwargs
                are needed.

        Returns:
            dict:
                Keyword arguments to use on the indicated model.
        """
        return copy.deepcopy(self._model_kwargs.get(model_name))

    def set_model_kwargs(self, model_name, model_kwargs):
        """Set the model kwargs used for the indicated model."""
        self._model_kwargs[model_name] = model_kwargs

    def _get_field_dtype(self, field_name, field_metadata):
        field_type = field_metadata['type']
        field_subtype = field_metadata.get('subtype')
        dtype = self._TYPES_TO_DTYPES.get((field_type, field_subtype))
        if not dtype:
            raise MetadataError(
                'Invalid type and subtype combination for field {}: ({}, {})'.format(
                    field_name, field_type, field_subtype)
            )

        return dtype

    def get_fields(self):
        """Get fields metadata.

        Returns:
            dict:
                Dictionary of fields metadata for this table.
        """
        return copy.deepcopy(self._fields_metadata)

    def get_dtypes(self, ids=False):
        """Get a ``dict`` with the ``dtypes`` for each field of the table.

        Args:
            ids (bool):
                Whether or not to include the id fields. Defaults to ``False``.

        Returns:
            dict:
                Dictionary that contains the field names and data types.
        """
        dtypes = dict()
        for name, field_meta in self._fields_metadata.items():
            field_type = field_meta['type']

            if ids or (field_type != 'id'):
                dtypes[name] = self._get_field_dtype(name, field_meta)

        return dtypes

    def _build_fields_metadata(self, data):
        """Build all the fields metadata.

        Args:
            data (pandas.DataFrame):
                Data to be analyzed.

        Returns:
            dict:
                Dict of valid fields.

        Raises:
            ValueError:
                If a column from the data analyzed is an unsupported data type
        """
        fields_metadata = dict()
        for field_name in self._field_names:
            if field_name not in data:
                raise ValueError('Field {} not found in given data'.format(field_name))

            field_meta = self._field_types.get(field_name)
            if field_meta:
                dtype = self._get_field_dtype(field_name, field_meta)
            else:
                dtype = data[field_name].dtype
                field_template = self._DTYPES_TO_TYPES.get(dtype.kind)
                if field_template is None:
                    msg = 'Unsupported dtype {} in column {}'.format(dtype, field_name)
                    raise ValueError(msg)

                field_meta = copy.deepcopy(field_template)

            field_transformer = self._field_transformers.get(field_name)
            if field_transformer:
                field_meta['transformer'] = field_transformer
            else:
                field_meta['transformer'] = self._dtype_transformers.get(np.dtype(dtype).kind)

            anonymize_category = self._anonymize_fields.get(field_name)
            if anonymize_category:
                field_meta['pii'] = True
                field_meta['pii_category'] = anonymize_category

            fields_metadata[field_name] = field_meta

        return fields_metadata

    def _get_transformers(self, dtypes):
        """Create the transformer instances needed to process the given dtypes.

        Args:
            dtypes (dict):
                mapping of field names and dtypes.

        Returns:
            dict:
                mapping of field names and transformer instances.
        """
        transformers = dict()
        for name, dtype in dtypes.items():
            field_metadata = self._fields_metadata.get(name, {})
            transformer_template = field_metadata.get('transformer')
            if transformer_template is None:
                transformer_template = self._dtype_transformers[np.dtype(dtype).kind]
                if transformer_template is None:
                    # Skip this dtype
                    continue

                field_metadata['transformer'] = transformer_template

            if isinstance(transformer_template, str):
                transformer_template = self._transformer_templates[transformer_template]

            if isinstance(transformer_template, type):
                transformer = transformer_template()
            else:
                transformer = copy.deepcopy(transformer_template)

            LOGGER.debug('Loading transformer %s for field %s',
                         transformer.__class__.__name__, name)
            transformers[name] = transformer

        return transformers

    def _fit_transform_constraints(self, data):
        for constraint in self._constraints:
            data = constraint.fit_transform(data)

        return data

    def _fit_hyper_transformer(self, data, extra_columns):
        """Create and return a new ``rdt.HyperTransformer`` instance.

        First get the ``dtypes`` and then use them to build a transformer dictionary
        to be used by the ``HyperTransformer``.

        Args:
            data (pandas.DataFrame):
                Data to transform.
            extra_columns (set):
                Names of columns that are not in the metadata but that should also
                be transformed. In most cases, these are the fields that were added
                by previous transformations which the data underwent.

        Returns:
            rdt.HyperTransformer
        """
        meta_dtypes = self.get_dtypes(ids=False)
        dtypes = {}
        numerical_extras = []
        for column in data.columns:
            if column in meta_dtypes:
                dtypes[column] = meta_dtypes[column]
            elif column in extra_columns:
                dtype_kind = data[column].dtype.kind
                if dtype_kind in ('i', 'f'):
                    numerical_extras.append(column)
                else:
                    dtypes[column] = dtype_kind

        transformers_dict = self._get_transformers(dtypes)
        for column in numerical_extras:
            transformers_dict[column] = rdt.transformers.NumericalTransformer()

        self._hyper_transformer = rdt.HyperTransformer(field_transformers=transformers_dict)
        self._hyper_transformer.fit(data[list(transformers_dict.keys())])

    @staticmethod
    def _get_key_subtype(field_meta):
        """Get the appropriate key subtype."""
        field_type = field_meta['type']

        if field_type == 'categorical':
            field_subtype = 'string'

        elif field_type in ('numerical', 'id'):
            field_subtype = field_meta['subtype']
            if field_subtype not in ('integer', 'string'):
                raise ValueError(
                    'Invalid field "subtype" for key field: "{}"'.format(field_subtype)
                )

        else:
            raise ValueError(
                'Invalid field "type" for key field: "{}"'.format(field_type)
            )

        return field_subtype

    def set_primary_key(self, primary_key):
        """Set the primary key of this table.

        The field must exist and either be an integer or categorical field.

        Args:
            primary_key (str or list):
                Name of the field(s) to be used as the new primary key.

        Raises:
            ValueError:
                If the table or the field do not exist or if the field has an
                invalid type or subtype.
        """
        if primary_key is not None:
            fields = primary_key if isinstance(primary_key, list) else [primary_key]
            for field_name in fields:
                if field_name not in self._fields_metadata:
                    raise ValueError('Field "{}" does not exist in this table'.format(field_name))

                field_metadata = self._fields_metadata[field_name]
                if field_metadata['type'] != 'id':
                    field_subtype = self._get_key_subtype(field_metadata)

                    field_metadata.update({
                        'type': 'id',
                        'subtype': field_subtype
                    })

        self._primary_key = primary_key

    def _make_anonymization_mappings(self, data):
        mappings = {}
        for name, field_metadata in self._fields_metadata.items():
            if field_metadata['type'] != 'id' and field_metadata.get('pii'):
                uniques = data[name].unique()
                mappings[name] = dict(
                    zip(uniques, Table._get_fake_values(field_metadata, len(uniques)))
                )

        self._ANONYMIZATION_MAPPINGS[id(self)] = mappings

    def _anonymize(self, data):
        anonymization_mappings = self._ANONYMIZATION_MAPPINGS.get(id(self))
        if anonymization_mappings:
            data = data.copy()
            for name, mapping in anonymization_mappings.items():
                if name in data:
                    data[name] = data[name].map(mapping)

        return data

    def fit(self, data):
        """Fit this metadata to the given data.

        Args:
            data (pandas.DataFrame):
                Table to be analyzed.
        """
        LOGGER.info('Fitting table %s metadata', self.name)
        if not self._field_names:
            self._field_names = list(data.columns)
        elif isinstance(self._field_names, set):
            self._field_names = [field for field in data.columns if field in self._field_names]

        self._dtypes = data[self._field_names].dtypes

        if not self._fields_metadata:
            self._fields_metadata = self._build_fields_metadata(data)

        # Re-set the primary key to validate its name and type
        self.set_primary_key(self._primary_key)

        self._make_anonymization_mappings(data)
        LOGGER.info('Anonymizing table %s', self.name)
        data = self._anonymize(data)

        LOGGER.info('Fitting constraints for table %s', self.name)
        constrained = self._fit_transform_constraints(data)
        extra_columns = set(constrained.columns) - set(data.columns)

        LOGGER.info('Fitting HyperTransformer for table %s', self.name)
        self._fit_hyper_transformer(constrained, extra_columns)
        self.fitted = True

    def _transform_constraints(self, data, on_missing_column='error'):
        for constraint in self._constraints:
            try:
                data = constraint.transform(data)
            except MissingConstraintColumnError:
                if on_missing_column == 'error':
                    raise MissingConstraintColumnError()

                elif on_missing_column == 'drop':
                    indices_to_drop = data.columns.isin(constraint.constraint_columns)
                    columns_to_drop = data.columns.where(indices_to_drop).dropna()
                    data = data.drop(columns_to_drop, axis=1)

                else:
                    raise ValueError('on_missing_column must be \'drop\' or \'error\'')

        return data

    def _validate_data_on_constraints(self, data):
        """Make sure the given data is valid for the given constraints.

        Args:
            data (pandas.DataFrame):
                Table data.

        Returns:
            None

        Raises:
            ConstraintsNotMetError:
                If the table data is not valid for the provided constraints.
        """
        for constraint in self._constraints:
            if set(constraint.constraint_columns).issubset(data.columns.values):
                if not constraint.is_valid(data).all():
                    raise ConstraintsNotMetError('Data is not valid for the given constraints')

    def transform(self, data, on_missing_column='error'):
        """Transform the given data.

        Args:
            data (pandas.DataFrame):
                Table data.
            on_missing_column (str):
                If the value is error, then a `MissingConstraintColumnError` is raised.
                If the value is drop, then the columns involved in the constraint that
                are present in data will be dropped.

        Returns:
            pandas.DataFrame:
                Transformed data.

        Raises:
            ConstraintsNotMetError:
                If the table data is not valid for the provided constraints.
        """
        if not self.fitted:
            raise MetadataNotFittedError()

        fields = [field for field in self.get_dtypes(ids=False) if field in data.columns]
        LOGGER.debug('Anonymizing table %s', self.name)
        data = self._anonymize(data[fields])

        self._validate_data_on_constraints(data)

        LOGGER.debug('Transforming constraints for table %s', self.name)
        data = self._transform_constraints(data, on_missing_column)

        LOGGER.debug('Transforming table %s', self.name)
        try:
            return self._hyper_transformer.transform(data)
        except rdt.errors.NotFittedError:
            return data

    @classmethod
    def _make_ids(cls, field_metadata, length):
        field_subtype = field_metadata.get('subtype', 'integer')
        if field_subtype == 'string':
            regex = field_metadata.get('regex', '[ABCDEFGHJKLMOPQRSTUWYZ]+')
            generator, max_size = strings_from_regex(regex)
            if max_size < length:
                raise ValueError((
                    'Unable to generate {} unique values for regex {}, the '
                    'maximum number of unique values is {}.'
                ).format(length, regex, max_size))
            values = [next(generator) for _ in range(length)]

            return pd.Series(list(values)[:length])
        else:
            return pd.Series(np.arange(length))

    def reverse_transform(self, data):
        """Reverse the transformed data to the original format.

        Args:
            data (pandas.DataFrame):
                Data to be reverse transformed.

        Returns:
            pandas.DataFrame
        """
        if not self.fitted:
            raise MetadataNotFittedError()

        try:
            reversed_data = self._hyper_transformer.reverse_transform(data)
        except rdt.errors.NotFittedError:
            reversed_data = data

        for constraint in reversed(self._constraints):
            reversed_data = constraint.reverse_transform(reversed_data)

        for name, field_metadata in self._fields_metadata.items():
            field_type = field_metadata['type']
            if field_type == 'id' and name not in reversed_data:
                field_data = self._make_ids(field_metadata, len(reversed_data))
            elif field_metadata.get('pii', False):
                field_data = pd.Series(Table._get_fake_values(field_metadata, len(reversed_data)))
            else:
                field_data = reversed_data[name]

            reversed_data[name] = field_data[field_data.notnull()].astype(self._dtypes[name])

        return reversed_data[self._field_names]

    def filter_valid(self, data):
        """Filter the data using the constraints and return only the valid rows.

        Args:
            data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Table containing only the valid rows.
        """
        for constraint in self._constraints:
            data = constraint.filter_valid(data)

        return data

    def make_ids_unique(self, data):
        """Repopulate any id fields in provided data to guarantee uniqueness.

        Args:
            data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Table where all id fields are unique.
        """
        for name, field_metadata in self._fields_metadata.items():
            if field_metadata['type'] == 'id' and not data[name].is_unique:
                ids = self._make_ids(field_metadata, len(data))
                ids.index = data.index.copy()
                data[name] = ids

        return data

    # ###################### #
    # Metadata Serialization #
    # ###################### #

    def to_dict(self):
        """Get a dict representation of this metadata.

        Returns:
            dict:
                dict representation of this metadata.
        """
        return {
            'fields': copy.deepcopy(self._fields_metadata),
            'constraints': [
                constraint if isinstance(constraint, dict) else constraint.to_dict()
                for constraint in self._constraints
            ],
            'model_kwargs': copy.deepcopy(self._model_kwargs),
            'name': self.name,
            'primary_key': self._primary_key,
            'sequence_index': self._sequence_index,
            'entity_columns': self._entity_columns,
            'context_columns': self._context_columns,
        }

    def to_json(self, path):
        """Dump this metadata into a JSON file.

        Args:
            path (str):
                Path of the JSON file where this metadata will be stored.
        """
        with open(path, 'w') as out_file:
            json.dump(self.to_dict(), out_file, indent=4)

    @classmethod
    def from_dict(cls, metadata_dict, dtype_transformers=None):
        """Load a Table from a metadata dict.

        Args:
            metadata_dict (dict):
                Dict metadata to load.
            dtype_transformers (dict):
                If passed, set the dtype_transformers on the new instance.
        """
        metadata_dict = copy.deepcopy(metadata_dict)
        fields = metadata_dict['fields'] or {}
        instance = cls(
            name=metadata_dict.get('name'),
            field_names=set(fields.keys()),
            field_types=fields,
            constraints=metadata_dict.get('constraints') or [],
            model_kwargs=metadata_dict.get('model_kwargs') or {},
            primary_key=metadata_dict.get('primary_key'),
            sequence_index=metadata_dict.get('sequence_index'),
            entity_columns=metadata_dict.get('entity_columns') or [],
            context_columns=metadata_dict.get('context_columns') or [],
            dtype_transformers=dtype_transformers,
            min_value=metadata_dict.get('min_value', 'auto'),
            max_value=metadata_dict.get('max_value', 'auto'),
            rounding=metadata_dict.get('rounding', 'auto'),
        )
        instance._fields_metadata = fields
        return instance

    @classmethod
    def from_json(cls, path):
        """Load a Table from a JSON.

        Args:
            path (str):
                Path of the JSON file to load
        """
        with open(path, 'r') as in_file:
            return cls.from_dict(json.load(in_file))



import pandas as pd
from scipy.stats import chi2_contingency
from sdmetrics.goal import Goal

from syntherela.metrics.base import SingleColumnMetric, StatisticalBaseMetric


class ChiSquareTest(StatisticalBaseMetric, SingleColumnMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "ChiSquareTest"
        self.goal = Goal.MINIMIZE

    @staticmethod
    def is_applicable(column_type):
        return column_type == "categorical" or column_type == "boolean"

    def validate(self, column):
        if column.dtype.name not in ("object", "category", "bool"):
            raise ValueError(
                f"{self.name} can only be applied to categorical columns, but column {column.name} is of type {column.dtype}"
            )

    @staticmethod
    def compute(real_data, synthetic_data):
        orig_col = pd.Categorical(real_data)
        synth_col = pd.Categorical(synthetic_data, categories=orig_col.categories)
        freq_orig = orig_col.value_counts()
        freq_synth = synth_col.value_counts()
        if freq_synth.sum() == 0:
            return {"statistic": -1, "p_value": 0}
        freq_synth = freq_synth / freq_synth.sum() * freq_orig.sum()
        assert (
            freq_orig.index == freq_synth.index
        ).all(), "Indexes do not match for column"
        # calculate the chi-square test
        statistic, pval, _, _ = chi2_contingency([freq_orig, freq_synth])

        return {"statistic": statistic, "p_value": pval}



import os
import argparse
from pathlib import Path
import glob
import sys

import pandas as pd
from realtabformer import REaLTabFormer
import logging
import wandb
from relsyndgb.metadata import Metadata
from relsyndgb.data import load_tables, save_tables, remove_sdv_columns

MODEL_NAME = "REALTABFORMER"
BATCH_SIZE_PARENT = 1024
BATCH_SIZE_CHILD = 1

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="rossmann_subsampled")
args.add_argument("--real_data_path", type=str, default="/d/hpc/projects/FRI/mj5835/relsyndgb/data/downloads")
args.add_argument("--synthetic_data_path", type=str, default="/d/hpc/projects/FRI/mj5835/relsyndgb/data/synthetic/")
args.add_argument("--full_sensitivity", type=bool, default=True)
args.add_argument("--retrain", type=bool, default=True)
args.add_argument("--run_id", type=str, default="1")
args = args.parse_args()
dataset_name = args.dataset_name
full_sensitivity = args.full_sensitivity
retrain = args.retrain
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
run_id = args.run_id

logger = logging.getLogger(f'{MODEL_NAME}_logger')

# logger.basicConfig(
#     level=logging.DEBUG,  # Set the desired log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
#     format='%(asctime)s - %(levelname)s - %(message)s',
#     # filename=f"{MODEL_NAME}_{dataset_name}.log",  # Specify the file where logs should be stored
#     stream=sys.stdout,
# )

logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.info(f"START LOGGING...")

os.environ["WANDB_PROJECT"]=f"{MODEL_NAME}_{dataset_name}"

# save your trained model checkpoint to wandb
os.environ["WANDB_LOG_MODEL"]="true"

# turn off watch to log faster
os.environ["WANDB_WATCH"]="false"

# os.environ["WANDB_DISABLED"] = "true"

os.environ["PYTORCH_CUDA_ALLOC_CONF"]= "expandable_segments:True"


logger.info(f"START {MODEL_NAME}...")


logger.debug("Loading real data...")
metadata = Metadata().load_from_json(Path(real_data_path) / f'{dataset_name}/metadata.json')
real_data = load_tables(Path(real_data_path) / f'{dataset_name}', metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata)
metadata.validate_data(real_data)
logger.debug("Real data loaded")

synthetic_data = {}

# GENERATE SYNTHETIC DATA ---------------------------------

root_table_name = metadata.get_root_tables()[0]
parent_df = real_data[root_table_name]

join_on = metadata.get_primary_key(root_table_name)
parent_model_path = f'models/{MODEL_NAME}/{dataset_name}/checkpoint_{root_table_name}' 

if retrain or not(os.path.exists(parent_model_path) and len(os.listdir(parent_model_path)) > 0):
    logger.debug("Starting parent table generation...")
    logger.debug("Fitting parent model...")
    parent_model = REaLTabFormer(model_type="tabular", batch_size=BATCH_SIZE_PARENT, report_to="wandb", epochs=5)
    os.makedirs(parent_model_path, exist_ok=True)
    # fit and save parent model
    parent_model.fit(parent_df.drop(join_on, axis=1), full_sensitivity=full_sensitivity)
    parent_model.save(parent_model_path)
    logger.debug("Parent fitted and saved")
else:
    logger.debug("Skipping parent table generation and reading last trained model...")

# load trained parent model
directories = list(filter(os.path.isdir, glob.glob(f"{parent_model_path}/id*")))
directories.sort(key=lambda x: os.path.getmtime(x))
parent_model_path = directories[-1]
parent_model = REaLTabFormer(
    model_type="tabular",
    parent_realtabformer_path=parent_model_path,
)

parent_model = parent_model.load_from_dir(parent_model_path)
# sample from parent model
logger.debug("Sampling from parent model...")
parent_samples = parent_model.sample(real_data[root_table_name].shape[0])
parent_samples.index.name = join_on
parent_samples = parent_samples.reset_index()
synthetic_data[root_table_name] = parent_samples
logger.debug("Done.")

#GENERATE CHILD TABLES
for child_table_name in metadata.get_children(root_table_name):
    logger.info(f"Starting child table generation for {child_table_name}...")
    child_model_path = f'models/{MODEL_NAME}/{dataset_name}/checkpoint_{child_table_name}'
    child_df = real_data[child_table_name]
    assert ((join_on in parent_df.columns) and
            (join_on in child_df.columns))
    if retrain or not(os.path.exists(child_model_path) and len(os.listdir(child_model_path)) > 0):
        child_model = REaLTabFormer(
            model_type="relational",
            parent_realtabformer_path=parent_model_path,
            output_max_length=None,
            train_size=0.95,
            batch_size=BATCH_SIZE_CHILD,
            report_to="wandb",
            epochs=5)
        logger.debug(f"Fitting child model for table {child_table_name}...")
        # fit child model
        child_model.fit(
            df = child_df,
            in_df=parent_df,
            join_on=join_on)
        # save child model
        os.makedirs(child_model_path, exist_ok=True)
        child_model.save(child_model_path)
        logger.debug("Child fitted and saved")

    directories = list(filter(os.path.isdir, glob.glob(f"{child_model_path}/id*")))
    directories.sort(key=lambda x: os.path.getmtime(x))
    child_model_path = directories[-1]
    child_model = REaLTabFormer(
        model_type="relational",
            parent_realtabformer_path=parent_model_path,
            output_max_length=None,
            train_size=0.95,
            batch_size=BATCH_SIZE_CHILD,
            report_to="wandb"
    )
    child_model = child_model.load_from_dir(child_model_path)
    
    logger.debug("Sampling from child model...")
    child_samples = child_model.sample(
        input_unique_ids=parent_samples[join_on],
        input_df=parent_samples.drop(join_on, axis=1),
        gen_batch=64)
    
    child_samples.index.name = join_on
    child_samples = child_samples.reset_index()
    
    synthetic_data[child_table_name] = child_samples
    logger.debug("Done.")


# SAVE DATA ---------------------------------
logger.info("Saving synthetic data...")
save_data_path = Path(synthetic_data_path) / dataset_name / MODEL_NAME / run_id
save_tables(synthetic_data, save_data_path)
logger.info("COMPLETE GENERATION DONE.")



import os

import pandas as pd
from sdv.datasets.demo import get_available_demos, download_demo


def load_tables(data_path, metadata):
    tables = {}
    for file_name in os.listdir(data_path):
        if not file_name.endswith(".csv"):
            continue
        table_name = file_name.split(".")[0]
        dtypes = {}
        parse_dates = []
        for column, column_info in metadata.tables[table_name].columns.items():
            if column_info["sdtype"] == "categorical":
                dtypes[column] = "category"
            elif column_info["sdtype"] == "boolean":
                dtypes[column] = "bool"
            elif column_info["sdtype"] == "datetime":
                parse_dates.append(column)
            # for ids and numerical values let pandas infer the type
        table = pd.read_csv(
            f"{data_path}/{file_name}",
            low_memory=False,
            dtype=dtypes,
            parse_dates=parse_dates,
        )
        tables[table_name] = table
    return tables


def remove_sdv_columns(tables, metadata, update_metadata=True, validate=True):
    """
    "_v1" Versions of the relational demo datasets in SDV have some columns that are not present in the original datasets.
    We created this function to remove these columns from the tables and the metadata.
    """
    for table_name, table in tables.items():
        for column in table.columns:
            if any(
                prefix in column
                for prefix in ["add_numerical", "nb_rows_in", "min(", "max(", "sum("]
            ):
                table = table.drop(columns=column, axis=1)

                if not update_metadata:
                    continue
                metadata.tables[table_name].columns.pop(column)

        tables[table_name] = table
    if validate:
        metadata.validate()
        metadata.validate_data(tables)
    return tables, metadata


def save_tables(tables, path):
    if not os.path.exists(path):
        os.makedirs(path)
    for table_name, table in tables.items():
        table.to_csv(os.path.join(path, f"{table_name}.csv"), index=False)


def download_sdv_relational_datasets(data_path="data/original"):
    """
    Download all the available relational datasets from SDV
    https://docs.sdv.dev/sdv/single-table-data/data-preparation/loading-data
    """
    sdv_relational_datasets = get_available_demos("multi_table")

    # iterate through the dataframe
    for dataset_name in sdv_relational_datasets.dataset_name:
        print(f"Downloading {dataset_name}...", end=" ")
        download_demo(
            "multi_table",
            dataset_name,
            output_folder_name=f"{data_path}/{dataset_name}",
        )
        print("Done.")


def denormalize_tables(tables, metadata):
    relationships = metadata.relationships.copy()
    denormalized_table = tables[relationships[0]["parent_table_name"]]
    already_merged_tables = [relationships[0]["parent_table_name"]]

    while len(relationships) > 0:
        if relationships[0]["parent_table_name"] in already_merged_tables:
            parent_table = denormalized_table
            child_table = tables[relationships[0]["child_table_name"]]
            already_merged_tables.append(relationships[0]["child_table_name"])

        elif relationships[0]["child_table_name"] in already_merged_tables:
            parent_table = tables[relationships[0]["parent_table_name"]]
            child_table = denormalized_table
            already_merged_tables.append(relationships[0]["parent_table_name"])
        else:
            relationships.append(relationships.pop(0))
            continue

        denormalized_table = parent_table.merge(
            child_table,
            left_on=relationships[0]["parent_primary_key"],
            right_on=relationships[0]["child_foreign_key"],
            suffixes=(None, f"_{relationships[0]['child_table_name']}"),
            how="outer",
        )

        # Drop the foreign key column with suffix from the denormalized table
        for column, column_info in metadata.tables[
            relationships[0]["child_table_name"]
        ].columns.items():
            if column_info["sdtype"] != "id":
                continue
            denormalized_table = drop_column_if_in_table(
                denormalized_table, f"{column}_{relationships[0]['child_table_name']}"
            )

        relationships.pop(0)

    return denormalized_table


def drop_column_if_in_table(table, column):
    if column in table.columns:
        table = table.drop(columns=column, axis=1)
    return table


def make_column_names_unique(real_data, synthetic_data, metadata, validate=True):
    for table_name in metadata.get_tables():
        if not real_data[table_name].columns.equals(synthetic_data[table_name].columns):
            raise ValueError("Real and synthetic data column names are not the same")

        table_metadata = metadata.tables[table_name].to_dict()

        for column in table_metadata["columns"]:
            real_data[table_name] = real_data[table_name].rename(
                columns={column: f"{table_name}_{column}"}
            )
            synthetic_data[table_name] = synthetic_data[table_name].rename(
                columns={column: f"{table_name}_{column}"}
            )
            metadata = metadata.rename_column(
                table_name, column, f"{table_name}_{column}"
            )

    if validate:
        metadata.validate()
        metadata.validate_data(real_data)
        metadata.validate_data(synthetic_data)

    return real_data, synthetic_data, metadata



import glob
import operator
import os
import platform
import re
import shutil
import stat
from pathlib import Path

from invoke import task


COMPARISONS = {
    '>=': operator.ge,
    '>': operator.gt,
    '<': operator.lt,
    '<=': operator.le
}

@task
def check_dependencies(c):
    c.run('python -m pip check')


@task
def unit(c):
    c.run('python -m pytest ./tests/unit --cov=rctgan --cov-report=xml')


@task
def integration(c):
    c.run('python -m pytest ./tests/integration --reruns 3')


def _validate_python_version(line):
    python_version_match = re.search(r"python_version(<=?|>=?)\'(\d\.?)+\'", line)
    if python_version_match:
        python_version = python_version_match.group(0)
        comparison = re.search(r'(>=?|<=?)', python_version).group(0)
        version_number = python_version.split(comparison)[-1].replace("'", "")
        comparison_function = COMPARISONS[comparison]
        return comparison_function(platform.python_version(), version_number)

    return True


@task
def install_minimum(c):
    with open('setup.py', 'r') as setup_py:
        lines = setup_py.read().splitlines()

    versions = []
    started = False
    for line in lines:
        if started:
            if line == ']':
                started = False
                continue

            line = line.strip()
            if _validate_python_version(line):
                requirement = re.match(r'[^>]*', line).group(0)
                requirement = re.sub(r"""['",]""", '', requirement)
                version = re.search(r'>=?[^(,|#)]*', line).group(0)
                if version:
                    version = re.sub(r'>=?', '==', version)
                    version = re.sub(r"""['",]""", '', version)
                    requirement += version
                versions.append(requirement)

        elif (line.startswith('install_requires = [') or
              line.startswith('pomegranate_requires = [')):
            started = True

    c.run(f'python -m pip install {" ".join(versions)}')


@task
def minimum(c):
    install_minimum(c)
    check_dependencies(c)
    unit(c)
    integration(c)


@task
def readme(c):
    test_path = Path('tests/readme_test')
    if test_path.exists() and test_path.is_dir():
        shutil.rmtree(test_path)

    cwd = os.getcwd()
    os.makedirs(test_path, exist_ok=True)
    shutil.copy('README.md', test_path / 'README.md')
    os.chdir(test_path)
    c.run('rundoc run --single-session python3 -t python3 README.md')
    os.chdir(cwd)
    shutil.rmtree(test_path)


@task
def tutorials(c):
    for ipynb_file in glob.glob('tutorials/*.ipynb') + glob.glob('tutorials/**/*.ipynb'):
        if '.ipynb_checkpoints' not in ipynb_file:
            c.run((
                'jupyter nbconvert --execute --ExecutePreprocessor.timeout=3600 '
                f'--to=html --stdout {ipynb_file}'
            ), hide='out')


@task
def lint(c):
    check_dependencies(c)
    c.run('flake8 rctgan')
    c.run('flake8 tests --ignore=D,SFS2')
    c.run('isort -c --recursive rctgan tests')
    c.run('pydocstyle rctgan')


def remove_readonly(func, path, _):
    "Clear the readonly bit and reattempt the removal"
    os.chmod(path, stat.S_IWRITE)
    func(path)


@task
def rmdir(c, path):
    try:
        shutil.rmtree(path, onerror=remove_readonly)
    except PermissionError:
        pass


# %%
import os
from dotenv import load_dotenv
from gretel_trainer.relational import *
from gretel_trainer.relational import RelationalData
import pandas as pd
from tqdm import tqdm
from IPython.display import display, HTML
from gretel_trainer.relational import MultiTable
from gretel_client import configure_session
from relsyndgb.metadata import Metadata
from relsyndgb.data import load_tables, save_tables, remove_sdv_columns
from pathlib import Path
import argparse

# %%
# get api key from .env
load_dotenv()

configure_session(api_key=os.environ.get("GRETEL_API_KEY"), validate=True)

DATASET_NAME = "rossmann-store-sales"
METHOD_NAME='gretel'
MODEL_NAME = "GRETEL"

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="rossmann_subsampled")
args.add_argument("--real_data_path", type=str, default="/d/hpc/projects/FRI/mj5835/relsyndgb/data/downloads")
args.add_argument("--synthetic_data_path", type=str, default="/d/hpc/projects/FRI/mj5835/relsyndgb/data/synthetic/")
args.add_argument("--run_id", type=str, default="1")
args = args.parse_args([])
dataset_name = args.dataset_name
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
run_id = args.run_id

logger = logging.getLogger(f'{MODEL_NAME}_logger')

logger.debug("Loading real data...")
metadata = Metadata().load_from_json(Path(real_data_path) / f'{dataset_name}/metadata.json')
real_data = load_tables(Path(real_data_path) / f'{dataset_name}', metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata)
metadata.validate_data(real_data)
logger.debug("Real data loaded")

synthetic_data = {}

# %%
# @title
# Alternatively, manually define relational data
# Uncomment code to run

csv_dir = real_data_path
parent_table = (f"store", "Store")
child1_table = (f"historical", "Id")
child1_table_fk = parent_table[1]
tables = [
    # ("table_name", "primary_key")
    parent_table,
    child1_table,
]

foreign_keys = [
    # ("fkey_table.fkey", "pkey_table.pkey")
    (f"{child1_table[0]}.{child1_table_fk}",
        f"{parent_table[0]}.{parent_table[1]}"),
]

relational_data = RelationalData()

# tables_train, tables_test = utils.get_train_test_split(DATASET_NAME, k)

for table, pk in tables:
    relational_data.add_table(
        name=table, primary_key=pk, data=real_data[table])

for fk, ref in foreign_keys:
    relational_data.add_foreign_key(foreign_key=fk, referencing=ref)

# print("\033[1m Source Data: \033[0m")
# source_data = join_tables(parent_table[0], child1_table[0], relational_data=relational_data)

gretel_model = "lstm"
multitable = MultiTable(
    relational_data,
    project_display_name=f"Synthesize {DATASET_NAME} - {gretel_model}",
    gretel_model=gretel_model,
    # refresh_interval=60
)
multitable.train()
multitable.generate(record_size_ratio=1)

table = "store"  # @param {type:"string"}

# source_table = multitable.relational_data.get_table_data(table).head(5)
# synth_table = multitable.synthetic_output_tables[table][source_table.columns].head(
#     5)
# print("\033[1m Source Table:")
# display(source_table)
# print("\n\n\033[1m Synthesized Table:")
# display(synth_table)

synthetic_data = multitable.synthetic_output_tables

logger.info("Saving synthetic data...")
save_data_path = Path(synthetic_data_path) / dataset_name / MODEL_NAME / run_id
save_tables(synthetic_data, save_data_path)
logger.info("COMPLETE GENERATION DONE.")
# utils.save_data(synthetic_data, DATASET_NAME, k, method=METHOD_NAME)

# %%
# METHOD_NAME='gretel'
# synthetic_data = multitable.synthetic_output_tables
# utils.save_data(synthetic_data, DATASET_NAME, k, method=METHOD_NAME)
# %%



import warnings

import xgboost as xgb
import seaborn as sns
from matplotlib import rc
import matplotlib.pyplot as plt

from syntherela.metadata import Metadata
from syntherela.metrics.multi_table.detection import AggregationDetection
from syntherela.data import load_tables, remove_sdv_columns

warnings.filterwarnings("ignore")
sns.set_theme()
rc("font", **{"family": "serif", "serif": ["Times"], "size": 30})
rc("text", usetex=True)


def reproduce_figure(
    tables, tables_synthetic, metadata, target_table, feature, figure_name
):
    # Compute the metric
    xgb_cls = xgb.XGBClassifier
    xgb_args = {
        "seed": 0,
    }

    metric = AggregationDetection(
        classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42
    )

    for table in tables.keys():
        tables_synthetic[table] = tables_synthetic[table][tables[table].columns]

    metric.run(
        tables,
        tables_synthetic,
        metadata=metadata,
        target_table=target_table,
    )

    # Plot the feature importance

    metric.plot_partial_dependence(feature, seed=0)
    plt.savefig(
        f"results/figures/figure5{figure_name}.png", bbox_inches="tight", dpi=600
    )


dataset_name = "imdb_MovieLens_v1"
method = "CLAVADDPM"

metadata = Metadata().load_from_json(f"data/original/{dataset_name}/metadata.json")

tables = load_tables(f"data/original/{dataset_name}/", metadata)
tables_synthetic = load_tables(
    f"data/synthetic/{dataset_name}/{method}/1/sample1", metadata
)

tables, metadata = remove_sdv_columns(tables, metadata)
tables_synthetic, metadata = remove_sdv_columns(
    tables_synthetic, metadata, update_metadata=False
)

## FIGURE 5 (a)
feature = "movies2actors_movieid_cast_num_nunique"
reproduce_figure(tables, tables_synthetic, metadata, "movies", feature, "a")
## FIGURE 5 (b)
feature = "u2base_movieid_rating_mean"
reproduce_figure(tables, tables_synthetic, metadata, "movies", feature, "b")



import os
import argparse
from pathlib import Path
import glob
import sys

import pandas as pd
from realtabformer import REaLTabFormer
import logging
import wandb
from relsyndgb.metadata import Metadata
from relsyndgb.data import load_tables, save_tables, remove_sdv_columns

MODEL_NAME = "REALTABFORMER"
BATCH_SIZE_PARENT = 1024
BATCH_SIZE_CHILD = 1

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="rossmann_subsampled")
args.add_argument("--real_data_path", type=str, default="/d/hpc/projects/FRI/mj5835/relsyndgb/data/downloads")
args.add_argument("--synthetic_data_path", type=str, default="/d/hpc/projects/FRI/mj5835/relsyndgb/data/synthetic/")
args.add_argument("--full_sensitivity", type=bool, default=True)
args.add_argument("--retrain", type=bool, default=True)
args.add_argument("--run_id", type=str, default="1")
args = args.parse_args()
dataset_name = args.dataset_name
full_sensitivity = args.full_sensitivity
retrain = args.retrain
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
run_id = args.run_id

logger = logging.getLogger(f'{MODEL_NAME}_logger')

# logger.basicConfig(
#     level=logging.DEBUG,  # Set the desired log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
#     format='%(asctime)s - %(levelname)s - %(message)s',
#     # filename=f"{MODEL_NAME}_{dataset_name}.log",  # Specify the file where logs should be stored
#     stream=sys.stdout,
# )

logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.info(f"START LOGGING...")

os.environ["WANDB_PROJECT"]=f"{MODEL_NAME}_{dataset_name}"

# save your trained model checkpoint to wandb
os.environ["WANDB_LOG_MODEL"]="true"

# turn off watch to log faster
os.environ["WANDB_WATCH"]="false"

# os.environ["WANDB_DISABLED"] = "true"

os.environ["PYTORCH_CUDA_ALLOC_CONF"]= "expandable_segments:True"


logger.info(f"START {MODEL_NAME}...")


logger.debug("Loading real data...")
metadata = Metadata().load_from_json(Path(real_data_path) / f'{dataset_name}/metadata.json')
real_data = load_tables(Path(real_data_path) / f'{dataset_name}', metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata)
metadata.validate_data(real_data)
logger.debug("Real data loaded")

synthetic_data = {}

# GENERATE SYNTHETIC DATA ---------------------------------


# SAVE DATA ---------------------------------
logger.info("Saving synthetic data...")
save_data_path = Path(synthetic_data_path) / dataset_name / MODEL_NAME / run_id
save_tables(synthetic_data, save_data_path)
logger.info("COMPLETE GENERATION DONE.")


import os
import argparse
from pathlib import Path
import glob
import sys

import pandas as pd
from rctgan import Metadata
from rctgan.relational import RCTGAN
import logging

MODEL_NAME = "RCTGAN"

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="rossmann_subsampled")
args.add_argument("--real_data_path", type=str, default="/d/hpc/projects/FRI/mj5835/relsyndgb/data/downloads")
args.add_argument("--synthetic_data_path", type=str, default="/d/hpc/projects/FRI/mj5835/relsyndgb/data/synthetic/")
args.add_argument("--full_sensitivity", type=bool, default=True)
args.add_argument("--retrain", type=bool, default=True)
args.add_argument("--run_id", type=str, default="1")
args = args.parse_args()
dataset_name = args.dataset_name
full_sensitivity = args.full_sensitivity
retrain = args.retrain
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
run_id = args.run_id

logger = logging.getLogger(f'{MODEL_NAME}_logger')

# logger.basicConfig(
#     level=logging.DEBUG,  # Set the desired log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
#     format='%(asctime)s - %(levelname)s - %(message)s',
#     # filename=f"{MODEL_NAME}_{dataset_name}.log",  # Specify the file where logs should be stored
#     stream=sys.stdout,
# )

logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.info(f"START LOGGING...")

os.environ["WANDB_PROJECT"]=f"{MODEL_NAME}_{dataset_name}"

# save your trained model checkpoint to wandb
os.environ["WANDB_LOG_MODEL"]="true"

# turn off watch to log faster
os.environ["WANDB_WATCH"]="false"

# os.environ["WANDB_DISABLED"] = "true"

os.environ["PYTORCH_CUDA_ALLOC_CONF"]= "expandable_segments:True"


logger.info(f"START {MODEL_NAME}...")

# utils
def load_tables(data_path, metadata):
    tables = {}
    for file_name in os.listdir(data_path):
        if not file_name.endswith('.csv'):
            continue
        table_name = file_name.split('.')[0]
        dtypes = {}
        parse_dates = []
        for column, column_info in metadata.to_dict()['tables'][table_name]['fields'].items():
            if column_info['type'] == 'categorical':
                dtypes[column] = 'object'
            elif column_info['type'] == 'boolean':
                dtypes[column] = 'bool'
            elif column_info['type'] == 'datetime':
                parse_dates.append(column)
            # for ids and numerical values let pandas infer the type
        table = pd.read_csv(f'{data_path}/{file_name}', low_memory=False, dtype=dtypes, parse_dates=parse_dates)
        tables[table_name] = table
    return tables

def remove_sdv_columns(tables, metadata, update_metadata=True):
    """
    "_v1" Versions of the relational demo datasets in SDV have some columns that are not present in the original datasets.
    We created this function to remove these columns from the tables and the metadata.
    We have also created the following issue in the SDV repo which adresses this problem: https://github.com/sdv-dev/SDV/issues/1776
    """
    for table_name, table in tables.items():
        for column in table.columns:
            if any(prefix in column for prefix in ['add_numerical', 'nb_rows_in', 'min(', 'max(', 'sum(']):
                table = table.drop(columns = column, axis=1)

                if not update_metadata:
                    continue
                metadata._metadata['tables'][table_name]['fields'].pop(column)

        tables[table_name] = table
    metadata.validate(tables)
    return tables, metadata


def save_tables(tables, path):
    if not os.path.exists(path):
        os.makedirs(path)
    for table_name, table in tables.items():
        table.to_csv(os.path.join(path, f"{table_name}.csv"), index=False)

# GENERATE SYNTHETIC DATA ---------------------------------
logger.debug("Loading real data...")
metadata = Metadata(metadata=str(Path(real_data_path) / f'{dataset_name}/metadata_v0.json'))
real_data = load_tables(Path(real_data_path) / f'{dataset_name}', metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata)
metadata.validate_data(real_data)
logger.debug("Real data loaded")

synthetic_data = {}

model = RCTGAN(metadata=metadata)
logger.info("Fitting model...")
model.fit(real_data)

logger.info("Sampling and saving synthetic data...")
for i in range(3):
    model.seed = i
    synthetic_data = model.sample()
    save_data_path = Path(synthetic_data_path) / dataset_name / MODEL_NAME / run_id / f"sample{i}"
    save_tables(synthetic_data, save_data_path)

# SAVE DATA ---------------------------------
# save_data_path = Path(synthetic_data_path) / dataset_name / MODEL_NAME / run_id
# save_tables(synthetic_data, save_data_path)
logger.info("COMPLETE GENERATION DONE.")


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script for the addons packages."""

import json
import os
import shutil
import sys
from copy import deepcopy
from glob import glob
from tempfile import TemporaryDirectory

from setuptools import find_namespace_packages, setup

import rctgan.rdt2

with open('README.md', encoding='utf-8') as readme_file:
    README = readme_file.read()

rdt2_VERSION = rctgan.rdt2.__version__
ADDONS_PATH = os.path.dirname(os.path.realpath(__file__))


def _build_setup(addon_json):

    with open(addon_json, 'r', encoding='utf-8') as f:
        addon = json.load(f)

    addon_name = addon.get('name')
    addon_module = addon.get('transformers')[0].split('.')
    addon_module = '.'.join(addon_module[:-2])

    install_requires = [f'rdt2>={rdt2_VERSION}']
    install_requires.extend(addon.get('requirements', []))

    # this does something
    setup(
        author='MIT Data To AI Lab',
        author_email='dailabmit@gmail.com',
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ],
        description='Reversible Data Transforms',
        include_package_data=False,
        install_requires=install_requires,
        keywords=['rctgan.rdt2', addon_name],
        license='MIT license',
        long_description=README,
        long_description_content_type='text/markdown',
        name=addon_name,
        packages=find_namespace_packages(include=[addon_module]),
        python_requires='>=3.6,<3.10',
        url='https://github.com/sdv-dev/rdt2',
        version=rdt2_VERSION,
        zip_safe=False,
    )


def _run():
    path = sys.argv[0]
    base_path = os.path.realpath(path).replace(path, '')

    # clear build if exists
    build_path = os.path.join(base_path, 'build')
    if os.path.exists(build_path):
        shutil.rmtree(build_path)

    families = deepcopy(sys.argv[1:])
    all_families = [family for family in os.listdir('.') if os.path.isdir(family)]

    families = list(set(families).intersection(set(all_families)))
    for addon in glob(f'{ADDONS_PATH}/*/*.json'):
        with TemporaryDirectory() as temp_dir:
            build_command = [
                path, 'bdist_wheel', '--keep-temp', '--dist-dir', 'dist', '--bdist-dir', temp_dir,
                'sdist', '--keep-temp', '--dist-dir', 'dist', 'egg_info', '--egg-base', temp_dir
            ]

            base_name = os.path.basename(os.path.dirname(addon))
            sys.argv = deepcopy(build_command)

            if not families:
                _build_setup(addon)
            else:
                if os.path.basename(os.path.dirname(addon)) in families:
                    _build_setup(addon)

            remove_addon_build = os.path.join(
                base_path, 'build', 'lib', 'rctgan.rdt2', 'transformers', 'addons', base_name
            )

            # delete only the processed addon folder
            shutil.rmtree(remove_addon_build)


_run()



import os
import re
import json
import warnings
from pathlib import Path

import numpy as np
import pandas as pd

warnings.filterwarnings("ignore")

datasets = [
    "airbnb-simplified_subsampled",
    "rossmann_subsampled",
    "walmart_subsampled",
    "Biodegradability_v1",
    "imdb_MovieLens_v1",
    "CORA_v1",
]

dataset_names_dict = {
    "airbnb-simplified_subsampled": "AirBnB",
    "rossmann_subsampled": "Rossmann",
    "walmart_subsampled": "Walmart",
    "Biodegradability_v1": "Biodeg.",
    "imdb_MovieLens_v1": "MovieLens",
    "CORA_v1": "CORA",
}

single_table_methods = [
    "bayesian_network",
    "ddpm",
    "ctgan",
    "nflow",
    "tvae",
]

all_methods = [
    "SDV",
    "RCTGAN",
    "REALTABFORMER",
    "MOSTLYAI",
    "GRETEL_ACTGAN",
    "GRETEL_LSTM",
    "CLAVADDPM",
] + single_table_methods

method_names_dict = {
    "SDV": "SDV",
    "RCTGAN": "RCTGAN",
    "REALTABFORMER": "REALTABF.",
    "MOSTLYAI": "MOSTLYAI",
    "GRETEL_ACTGAN": "G-ACTGAN",
    "GRETEL_LSTM": "G-LSTM",
    "CLAVADDPM": "CLAVADDPM",
    "ddpm": "DDPM",
    "ctgan": "CTGAN",
    "nflow": "NFLOW",
    "tvae": "TVAE",
    "bayesian_network": "BN",
}

dataset_method_dict = {
    "airbnb-simplified_subsampled": [
        "SDV",
        "RCTGAN",
        "REALTABFORMER",
        "MOSTLYAI",
        "GRETEL_ACTGAN",
        "GRETEL_LSTM",
        "CLAVADDPM",
        "bayesian_network",
        "ctgan",
        "ddpm",
        "nflow",
        "tvae",
    ],
    "rossmann_subsampled": [
        "SDV",
        "RCTGAN",
        "REALTABFORMER",
        "MOSTLYAI",
        "GRETEL_ACTGAN",
        "GRETEL_LSTM",
        "CLAVADDPM",
        "bayesian_network",
        "ctgan",
        "ddpm",
        "nflow",
        "tvae",
    ],
    "walmart_subsampled": [
        "SDV",
        "RCTGAN",
        "REALTABFORMER",
        "MOSTLYAI",
        "GRETEL_ACTGAN",
        "GRETEL_LSTM",
        "CLAVADDPM",
        "bayesian_network",
        "ctgan",
        "ddpm",
        "nflow",
        "tvae",
    ],
    "Biodegradability_v1": [
        "SDV",
        "RCTGAN",
        "MOSTLYAI",
        "GRETEL_ACTGAN",
        "GRETEL_LSTM",
        "bayesian_network",
        "ctgan",
        "ddpm",
        "nflow",
        "tvae",
    ],
    "imdb_MovieLens_v1": ["RCTGAN", "MOSTLYAI", "GRETEL_ACTGAN", "CLAVADDPM", "ddpm"],
    "CORA_v1": [
        "SDV",
        "RCTGAN",
        "GRETEL_ACTGAN",
        "GRETEL_LSTM",
        "bayesian_network",
        "ctgan",
        "ddpm",
        "nflow",
        "tvae",
    ],
}

RESULTS_PATH = Path("./results")
RUNS = ["1", "2", "3"]

ALPHA_STATISTICAL = 0.05
ALPHA_DETECTION = 0.05

INCLUDE_SINGLE_TABLE = True


def get_methods(dataset, metric_type):
    methods = dataset_method_dict[dataset]
    if metric_type == "multi_table_metrics" or not INCLUDE_SINGLE_TABLE:
        methods = [method for method in methods if method not in single_table_methods]
    return methods


def create_table(metric_type, table_name="table1"):
    all_results = {}
    for run_id in RUNS:
        all_results[run_id] = {}
        for dataset in datasets:
            all_results[run_id][dataset] = {}
            methods = get_methods(dataset, metric_type)
            # take only methods that are in all_methods
            methods = [method for method in methods if method in all_methods]
            for method in methods:
                for file in os.listdir(f"{RESULTS_PATH}/{run_id}"):
                    if file.startswith(f"{dataset}_{method}"):
                        with open(RESULTS_PATH / run_id / file, "r") as f:
                            all_results[run_id][dataset][method] = json.load(f)

    base_metrics = []

    if metric_type == "single_column_metrics":
        base_metrics = [
            "ChiSquareTest",
            "KolmogorovSmirnovTest",
            "HellingerDistance",
            "JensenShannonDistance",
            "TotalVariationDistance",
            "WassersteinDistance",
            "SingleColumnDetection-LogisticRegression",
            "SingleColumnDetection-XGBClassifier",
        ]

    if metric_type == "single_table_metrics":
        base_metrics = [
            "MaximumMeanDiscrepancy",
            "PairwiseCorrelationDifference",
            "SingleTableDetection-LogisticRegression",
            "SingleTableDetection-XGBClassifier",
        ]

    if metric_type == "multi_table_metrics":
        base_metrics = [
            "CardinalityShapeSimilarity",
            "AggregationDetection-LogisticRegression",
            "AggregationDetection-XGBClassifier",
        ]

    df_run_list = []
    df_copy_list = []
    df_how_many_runs = pd.DataFrame(columns=["dataset", "methods"] + base_metrics)
    for run_id in RUNS:
        df = pd.DataFrame(columns=["dataset", "methods"] + base_metrics)
        df_latex = df.copy()
        for idx, dataset in enumerate(datasets):
            cnt = True
            methods = get_methods(dataset, metric_type)
            for method in methods:
                # append row with dataset, method and empty values for all metrics
                df = pd.concat(
                    [
                        df,
                        pd.DataFrame(
                            [[dataset, method] + [""] * len(base_metrics)],
                            columns=["dataset", "methods"] + base_metrics,
                        ),
                    ],
                    ignore_index=True,
                )
                df_how_many_runs = pd.concat(
                    [
                        df_how_many_runs,
                        pd.DataFrame(
                            [[dataset, method] + [""] * len(base_metrics)],
                            columns=["dataset", "methods"] + base_metrics,
                        ),
                    ],
                    ignore_index=True,
                )
                dataset_latex_text = f"\\multirow{{{len(dataset_method_dict[dataset])}}}{{*}}{{{dataset_names_dict[dataset]}}}"
                df_latex = pd.concat(
                    [
                        df_latex,
                        pd.DataFrame(
                            [
                                [
                                    dataset_latex_text if cnt else "",
                                    method_names_dict[method],
                                ]
                                + [""] * len(base_metrics)
                            ],
                            columns=["dataset", "methods"] + base_metrics,
                        ),
                    ],
                    ignore_index=True,
                )
                cnt = False

        df["AGG"] = 0
        df_how_many_runs["AGG"] = 0

        df_copy = df.copy()

        for dataset in datasets:
            detection_dict = {
                "SDV": {},
                "RCTGAN": {},
                "REALTABFORMER": {},
                "MOSTLYAI": {},
                "GRETEL_ACTGAN": {},
                "GRETEL_LSTM": {},
                "CLAVADDPM": {},
                "bayesian_network": {},
                "ddpm": {},
                "ctgan": {},
                "nflow": {},
                "tvae": {},
            }
            if metric_type == "single_column_metrics":
                for metric in base_metrics:
                    methods = get_methods(dataset, metric_type)
                    for method in methods:
                        if (
                            metric
                            not in all_results[run_id][dataset][method][metric_type]
                        ):
                            continue
                        for table in all_results[run_id][dataset][method][metric_type][
                            metric
                        ]:
                            for column in all_results[run_id][dataset][method][
                                metric_type
                            ][metric][table]:
                                detection_dict[method].setdefault(metric, {})
                                detection_dict[method][metric].setdefault("detected", 0)
                                detection_dict[method][metric].setdefault("all", 0)
                                detection_dict[method][metric].setdefault("copied", 0)
                                detection_dict[method][metric]["all"] += 1

                                if (
                                    "p_value"
                                    in all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]
                                ):
                                    metric_value = all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]["p_value"]
                                    if metric_value <= ALPHA_STATISTICAL:
                                        detection_dict[method][metric]["detected"] += 1
                                if (
                                    "p_val"
                                    in all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]
                                ):
                                    metric_value = all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]["p_val"]
                                    if metric_value <= ALPHA_STATISTICAL:
                                        detection_dict[method][metric]["detected"] += 1
                                if (
                                    "pval"
                                    in all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]
                                ):
                                    metric_value = all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]["pval"]
                                    if metric_value <= ALPHA_STATISTICAL:
                                        detection_dict[method][metric]["detected"] += 1
                                if (
                                    "value"
                                    in all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]
                                ):
                                    metric_value = all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]["value"]
                                    lower_bound = all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]["reference_ci"][0]
                                    upper_bound = all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]["reference_ci"][1]
                                    if (
                                        metric_value < lower_bound
                                        or metric_value > upper_bound
                                    ):
                                        detection_dict[method][metric]["detected"] += 1
                                if (
                                    "bin_test_p_val"
                                    in all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]
                                ):
                                    metric_value = all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]["bin_test_p_val"]
                                    if metric_value <= ALPHA_DETECTION:
                                        detection_dict[method][metric]["detected"] += 1
                                if (
                                    "copying_p_val"
                                    in all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]
                                ):
                                    metric_value = all_results[run_id][dataset][method][
                                        metric_type
                                    ][metric][table][column]["copying_p_val"]
                                    if metric_value <= ALPHA_DETECTION:
                                        detection_dict[method][metric]["copied"] += 1

            if metric_type == "single_table_metrics":
                for metric in base_metrics:
                    methods = get_methods(dataset, metric_type)
                    for method in methods:
                        if (
                            metric
                            not in all_results[run_id][dataset][method][metric_type]
                        ):
                            continue
                        for table in all_results[run_id][dataset][method][metric_type][
                            metric
                        ]:
                            detection_dict[method].setdefault(metric, {})
                            detection_dict[method][metric].setdefault("detected", 0)
                            detection_dict[method][metric].setdefault("all", 0)
                            detection_dict[method][metric].setdefault("copied", 0)
                            detection_dict[method][metric]["all"] += 1
                            if (
                                "p_value"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["p_value"]
                                if metric_value <= ALPHA_STATISTICAL:
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "p_val"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["p_val"]
                                if metric_value <= ALPHA_STATISTICAL:
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "pval"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["pval"]
                                if metric_value <= ALPHA_STATISTICAL:
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "value"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["value"]
                                lower_bound = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["reference_ci"][0]
                                upper_bound = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["reference_ci"][1]
                                if (
                                    metric_value < lower_bound
                                    or metric_value > upper_bound
                                ):
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "bin_test_p_val"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["bin_test_p_val"]
                                if metric_value <= ALPHA_DETECTION:
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "copying_p_val"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["copying_p_val"]
                                if metric_value <= ALPHA_DETECTION:
                                    detection_dict[method][metric]["copied"] += 1

            if metric_type == "multi_table_metrics":
                for metric in base_metrics:
                    methods = get_methods(dataset, metric_type)
                    for method in methods:
                        if (
                            metric
                            not in all_results[run_id][dataset][method][metric_type]
                        ):
                            continue
                        for table in all_results[run_id][dataset][method][metric_type][
                            metric
                        ]:
                            detection_dict[method].setdefault(metric, {})
                            detection_dict[method][metric].setdefault("detected", 0)
                            detection_dict[method][metric].setdefault("all", 0)
                            detection_dict[method][metric].setdefault("copied", 0)
                            detection_dict[method][metric]["all"] += 1
                            if (
                                "p_value"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["p_value"]
                                if metric_value <= ALPHA_STATISTICAL:
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "p_val"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["p_val"]
                                if metric_value <= ALPHA_STATISTICAL:
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "pval"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["pval"]
                                if metric_value <= ALPHA_STATISTICAL:
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "value"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["value"]
                                lower_bound = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["reference_ci"][0]
                                upper_bound = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["reference_ci"][1]
                                if (
                                    metric_value < lower_bound
                                    or metric_value > upper_bound
                                ):
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "bin_test_p_val"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["bin_test_p_val"]
                                if metric_value <= ALPHA_DETECTION:
                                    detection_dict[method][metric]["detected"] += 1
                            if (
                                "copying_p_val"
                                in all_results[run_id][dataset][method][metric_type][
                                    metric
                                ][table]
                            ):
                                metric_value = all_results[run_id][dataset][method][
                                    metric_type
                                ][metric][table]["copying_p_val"]
                                if metric_value <= ALPHA_DETECTION:
                                    detection_dict[method][metric]["copied"] += 1

            for method in dataset_method_dict[dataset]:
                for metric in base_metrics:
                    if metric not in detection_dict[method]:
                        continue
                    df.loc[
                        (df["methods"] == method) & (df["dataset"] == dataset), metric
                    ] = detection_dict[method][metric]["detected"]
                    df_copy.loc[
                        (df_copy["methods"] == method)
                        & (df_copy["dataset"] == dataset),
                        metric,
                    ] = detection_dict[method][metric]["copied"]
                    df_how_many_runs.loc[
                        (df_how_many_runs["methods"] == method)
                        & (df_how_many_runs["dataset"] == dataset),
                        metric,
                    ] = detection_dict[method][metric]["all"]
                    df.loc[
                        (df["methods"] == method) & (df["dataset"] == dataset), "AGG"
                    ] += detection_dict[method][metric]["detected"]
                    df_how_many_runs.loc[
                        (df_how_many_runs["methods"] == method)
                        & (df_how_many_runs["dataset"] == dataset),
                        "AGG",
                    ] += detection_dict[method][metric]["all"]

        # set all columns except dataset and methods to be float
        # fill empty values with nan
        for col in df.columns[2:]:
            df[col] = df[col].replace("", np.nan)
            df[col] = df[col].astype(float)
        df_run_list.append(df.copy())
        df_copy_list.append(df_copy.copy())

    df_average = (
        pd.concat(df_run_list)
        .groupby(["dataset", "methods"], sort=False)
        .mean()
        .reset_index()
    )
    df_sum = (
        pd.concat(df_run_list)
        .groupby(["dataset", "methods"], sort=False)
        .sum()
        .reset_index()
    )

    def assign_multicolumn_string(average, se):
        return f"\\multicolumn {{1}}{{c}}{{${average:.2f}\\pm{se:.3f}$}}"

    # iterate all rows and columns
    for i in range(df_latex.shape[0]):
        for j in range(df_latex.shape[1]):
            if j < 2:
                continue

            if df_average.iloc[i, j] != df_average.iloc[i, j]:
                df_latex.iloc[i, j] = "\multicolumn {1}{c}{-}"
                continue
            temp_result_string = ""
            for idx_, df_temp in enumerate(df_run_list):
                temp_result_string += f"{int(df_temp.iloc[i, j])}"
                temp_result_string += ", "

            temp_result_string = temp_result_string[:-2]

            temp_result_latex = f"{temp_result_string} ({df_how_many_runs.iloc[i, j]})"

            df_latex.iloc[i, j] = f"\\multicolumn {{1}}{{c}}{{{temp_result_latex}}}"

    for dataset in datasets:
        for metric in base_metrics:
            metric_results = df_average.loc[(df_average["dataset"] == dataset), metric]
            # get the lowest number of column [0]
            metric_results = metric_results.apply(pd.to_numeric)

            min_indexes = []

            # Iterate over columns
            min_val = metric_results.min()
            for idx, value in metric_results.items():
                if metric_results[idx] == min_val:
                    min_indexes.append(idx)

            min_indexes
            if len(min_indexes) == len(dataset_method_dict[dataset]):
                continue
            for idx in min_indexes:
                temp_result_list = []
                for df_temp in df_run_list:
                    temp_result_list.append(int(df_temp.loc[idx, metric]))

                temp_result_string = ""
                for idx_, df_temp in enumerate(df_run_list):
                    temp_result_string += f"{int(df_temp.loc[idx, metric])}"
                    temp_result_string += ", "

                temp_result_string = temp_result_string[:-2]
                temp_result_latex = (
                    f"{temp_result_string} ({df_how_many_runs.loc[idx, metric]})"
                )

                df_latex.loc[idx, metric] = (
                    f"\\multicolumn {{1}}{{c}}{{\\textbf{{{temp_result_latex}}}}}"
                )

    for dataset in datasets:
        agg_res = df_sum.loc[(df_sum["dataset"] == dataset), ["AGG"]]

        min_indexes = []

        min_val = agg_res["AGG"].min()
        for idx, row in agg_res.iterrows():
            if agg_res["AGG"][idx] == min_val:
                min_indexes.append(idx)

        min_indexes
        if len(min_indexes) == len(dataset_method_dict[dataset]):
            continue
        for idx in min_indexes:
            df_latex.loc[idx, "methods"] = (
                f"\\multirow{{1}}{{*}}{{\\textbf{{{method_names_dict[df_sum.loc[idx, 'methods']]}}}}}"
            )

    latex_table = df_latex.to_latex(header=False, index=False, escape=False)
    cline = f"\\ \cline{{2-{len(base_metrics) + 2}}}\n"
    latex_table = latex_table.replace("\\\n", cline)

    # iterate latex_table string row by row
    latex_table = latex_table.split("\n")

    for idx, row in enumerate(latex_table):
        # if row starts with multirow
        if row.lstrip().startswith("\multirow"):
            # remove cline from previous row
            latex_table[idx - 1] = latex_table[idx - 1].replace(cline[:-1], "\\ \hline")
            # add \n to previous row
            latex_table[idx - 1] += "\n"

    latex_table[-4] = latex_table[-4].replace(cline[:-1], "\\ \hline") + "\n"

    # join all rows back to one string
    latex_table = "\n".join(latex_table)

    latex_table = latex_table.replace("\\textbackslash ", "\\")

    table_latex = re.sub(" +", " ", latex_table)
    # save the latex table to file
    print(f"Saving Table {table_name}.")
    with open(f"results/tables/table{table_name}.tex", "w") as f:
        f.write(table_latex)


create_table("multi_table_metrics", "1")
create_table("single_column_metrics", "9")
create_table("single_table_metrics", "10")



import pandas as pd
from sdmetrics.goal import Goal
from sdmetrics.utils import is_datetime
from scipy.stats import wasserstein_distance

from syntherela.metrics.base import SingleColumnMetric, DistanceBaseMetric


class WassersteinDistance(DistanceBaseMetric, SingleColumnMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "WassersteinDistance"
        self.goal = Goal.MINIMIZE
        self.min_value = 0.0
        self.max_value = float("inf")

    @staticmethod
    def is_applicable(column_type):
        return column_type in ["numerical", "datetime"]

    @staticmethod
    def compute(real_data, synthetic_data, xmin, xmax, **kwargs):
        orig_col = pd.Series(real_data).dropna()
        synth_col = pd.Series(synthetic_data).dropna()
        # sample real and synthetic data to have the same length
        n = min(len(orig_col), len(synth_col))
        orig_col = orig_col.sample(n, random_state=0)
        synth_col = synth_col.sample(n, random_state=1)

        # scale data to [0, 1]
        x_orig = ((orig_col - xmin) / (xmax - xmin)).values
        x_synth = ((synth_col - xmin) / (xmax - xmin)).values
        return wasserstein_distance(x_orig, x_synth)

    def run(self, real_data, synthetic_data, **kwargs):
        if is_datetime(real_data):
            real_data = pd.to_numeric(real_data, errors="coerce", downcast="integer")
            synthetic_data = pd.to_numeric(
                synthetic_data, errors="coerce", downcast="integer"
            )
        combined_data = pd.concat([real_data, synthetic_data], ignore_index=True)
        xmin, xmax = combined_data.min(), combined_data.max()
        if self.is_constant(real_data):
            return {
                "value": 0,
                "reference_ci": [0, 0],
                "bootstrap_mean": 0,
                "bootstrap_se": 0,
            }
        return super().run(real_data, synthetic_data, xmin=xmin, xmax=xmax, **kwargs)



import pandas as pd
import seaborn as sns
from matplotlib import rc
import matplotlib.pyplot as plt

from syntherela.metadata import Metadata
from syntherela.data import load_tables, remove_sdv_columns

sns.set_theme()
rc("font", **{"family": "serif", "serif": ["Times"], "size": 30})
rc("text", usetex=True)

methods = [
    "SDV",
    "RCTGAN",
    "REALTABFORMER",
    "MOSTLYAI",
    "GRETEL_ACTGAN",
    "GRETEL_LSTM",
    "CLAVADDPM",
]


def prettify_method_name(method_name):
    if method_name == "REALTABFORMER":
        return "REALTABF."
    if method_name == "GRETEL_ACTGAN":
        return "G-ACTGAN"
    if method_name == "GRETEL_LSTM":
        return "G-LSTM"
    return method_name


# Load original data
metadata = Metadata().load_from_json("data/original/rossmann_subsampled/metadata.json")
tables = load_tables("data/original/rossmann_subsampled/", metadata)

tables, metadata = remove_sdv_columns(tables, metadata)

# Load synthetic data for all methods
all_tables = dict()
for method in methods:
    method_name = prettify_method_name(method)
    all_tables[method_name] = load_tables(
        f"data/synthetic/rossmann_subsampled/{method}/1/sample1", metadata
    )
    all_tables[method_name], metadata = remove_sdv_columns(
        all_tables[method_name], metadata, update_metadata=False
    )

COLORMAP = [
    "#377eb8",
    "#ff7f00",
    "#4daf4a",
    "#f781bf",
    "#a65628",
    "#984ea3",
    "#999999",
    "#e41a1c",
    "#dede00",
]

COLOR_DICT = {
    "REAL": COLORMAP[6],
    "MOSTLYAI": COLORMAP[8],
    "SDV": COLORMAP[0],
    "RCTGAN": COLORMAP[1],
    "REALTABF.": COLORMAP[2],
    "G-ACTGAN": COLORMAP[3],
    "G-LSTM": COLORMAP[5],
    "CLAVADDPM": COLORMAP[4],
}

fig, axes = plt.subplots(1, 2, figsize=(15, 5))

all_tables = dict()
for method in methods:
    method_name = prettify_method_name(method)
    all_tables[method_name] = load_tables(
        f"data/synthetic/rossmann_subsampled/{method}/1/sample1", metadata
    )
    all_tables[method_name], metadata = remove_sdv_columns(
        all_tables[method_name], metadata, update_metadata=False
    )

table = "store"
column1 = "PromoInterval"
columns = [
    tables[table][column1],
]
for method in methods:
    method = prettify_method_name(method)
    columns.append(all_tables[method][table][column1])
data = pd.DataFrame(pd.concat(columns, axis=0)).astype("object")
data.fillna("missing", inplace=True)


# map the 'Jan,Apr,Jul,Oct' to 'J,A,O'; 'Feb,May,Aug,Nov' to 'F,M,A,N'; 'Mar,Jun,Sep,Dec' to 'M,J,S,D'
data[column1] = (
    data[column1]
    .astype("object")
    .map(
        {
            "Jan,Apr,Jul,Oct": "J-A-J-O",
            "Feb,May,Aug,Nov": "F-M-A-N",
            "Mar,Jun,Sept,Dec": "M-J-S-D",
            "missing": "missing",
        }
    )
)

# sort the data by the frequency of the values in the real data
column_order = (
    tables[table][column1]
    .astype("object")
    .fillna("missing")
    .map(
        {
            "Jan,Apr,Jul,Oct": "J-A-J-O",
            "Feb,May,Aug,Nov": "F-M-A-N",
            "Mar,Jun,Sept,Dec": "M-J-S-D",
            "missing": "missing",
        }
    )
    .value_counts()
    .index.tolist()
)
# map the column order to the data

data[column1] = pd.Categorical(data[column1], categories=column_order, ordered=True)

methods_column = ["REAL"] * len(tables[table])
for method in methods:
    method = prettify_method_name(method)
    methods_column = methods_column + [method] * len(all_tables[method][table])
data["Data"] = methods_column

sns.histplot(
    data=data.dropna(),
    x=column1,
    hue="Data",
    multiple="dodge",
    stat="density",
    common_norm=False,
    legend=True,
    ax=axes[0],
    shrink=0.8,
    palette=COLOR_DICT,
    alpha=0.8,
)

table = "historical"
column2 = "Customers"
columns = [
    tables[table][column2],
]
for method in methods:
    method = prettify_method_name(method)
    columns.append(all_tables[method][table][column2])
data = pd.DataFrame(pd.concat(columns, axis=0))

methods_column = ["REAL"] * len(tables[table])
for method in methods:
    method = prettify_method_name(method)
    methods_column = methods_column + [method] * len(all_tables[method][table])
data["Data"] = methods_column

data.dropna(inplace=True)
sns.kdeplot(
    data=data,
    x=column2,
    hue="Data",
    common_norm=False,
    fill=False,
    legend=True,
    ax=axes[1],
    clip=(0, data[column2].max()),
    palette=COLOR_DICT,
    alpha=0.8,
)

fig.suptitle = f"{table}"

axes[0].tick_params(axis="x", labelsize=16)
axes[0].tick_params(axis="y", labelsize=16)
axes[0].set_ylabel("Density", fontsize=20)
axes[0].set_xlabel("PromoInterval", fontsize=20)
plt.setp(axes[0].get_legend().get_texts(), fontsize="16")
plt.setp(axes[0].get_legend().get_title(), fontsize="16")

axes[1].tick_params(axis="x", labelsize=16)
axes[1].tick_params(axis="y", labelsize=16)
axes[1].set_ylabel("Density", fontsize=20)
axes[1].set_xlabel("Customers", fontsize=20)
plt.setp(axes[1].get_legend().get_texts(), fontsize="16")
plt.setp(axes[1].get_legend().get_title(), fontsize="16")

fig.tight_layout()

plt.savefig("results/figures/figure1.png", dpi=600)



import pandas as pd
from sdv.metadata import MultiTableMetadata


class Metadata(MultiTableMetadata):
    def __init__(self, dataset_name=""):
        super().__init__()
        self.dataset_name = dataset_name

    def get_tables(self):
        return list(self.tables.keys())

    def get_primary_key(self, table_name):
        return self.tables[table_name].primary_key

    def get_table_meta(self, table_name, to_dict=True):
        table_meta = self.tables[table_name]
        if to_dict:
            return table_meta.to_dict()
        return table_meta

    def get_children(self, table_name):
        children = set()
        for relation in self.relationships:
            if relation["parent_table_name"] == table_name:
                children.add(relation["child_table_name"])
        return children

    def get_parents(self, table_name):
        parents = set()
        for relation in self.relationships:
            if relation["child_table_name"] == table_name:
                parents.add(relation["parent_table_name"])
        return parents

    def get_foreign_keys(self, parent_table_name, child_table_name):
        return self._get_foreign_keys(parent_table_name, child_table_name)

    def rename_column(self, table_name, old_column_name, new_column_name):
        self.tables[table_name].columns[new_column_name] = self.tables[
            table_name
        ].columns.pop(old_column_name)
        if self.tables[table_name].columns[new_column_name]["sdtype"] != "id":
            return self

        if self.tables[table_name].primary_key == old_column_name:
            self.tables[table_name].primary_key = new_column_name

        for relationship in self.relationships:
            if (
                relationship["parent_table_name"] == table_name
                and relationship["parent_primary_key"] == old_column_name
            ):
                relationship["parent_primary_key"] = new_column_name
            if (
                relationship["child_table_name"] == table_name
                and relationship["child_foreign_key"] == old_column_name
            ):
                relationship["child_foreign_key"] = new_column_name
        return self

    def get_root_tables(self):
        root_tables = set(self.tables.keys())
        for relation in self.relationships:
            root_tables.discard(relation["child_table_name"])
        return list(root_tables)


def drop_ids(table: pd.DataFrame, metadata: dict):
    for column, column_info in metadata["columns"].items():
        if column_info["sdtype"] == "id" and column in table.columns:
            table = table.drop(columns=column, axis=1)
    return table


def convert_metadata_to_v0(metadata):
    metadata_v1 = metadata.to_dict()
    metadata_v0 = {"tables": {}}
    for table_name, table_info in metadata_v1["tables"].items():
        metadata_v0["tables"][table_name] = {"fields": {}}
        for column, column_info in table_info["columns"].items():
            metadata_v0["tables"][table_name]["fields"][column] = {
                "type": column_info["sdtype"]
            }
            if column_info["sdtype"] == "boolean":
                # convert boolean to categorical
                metadata_v0["tables"][table_name]["fields"][column]["type"] = (
                    "categorical"
                )
            if column_info["sdtype"] == "datetime":
                metadata_v0["tables"][table_name]["fields"][column]["format"] = (
                    column_info["datetime_format"]
                )

        if "primary_key" in table_info:
            metadata_v0["tables"][table_name]["fields"][table_info["primary_key"]] = {
                "type": "id",
                "subtype": "string",
            }
            metadata_v0["tables"][table_name]["primary_key"] = table_info["primary_key"]

    for relationship in metadata_v1["relationships"]:
        parent_table_name = relationship["parent_table_name"]
        child_table_name = relationship["child_table_name"]
        parent_primary_key = relationship["parent_primary_key"]
        child_foreign_key = relationship["child_foreign_key"]
        metadata_v0["tables"][child_table_name]["fields"][child_foreign_key]["ref"] = {
            "table": parent_table_name,
            "field": parent_primary_key,
        }
        metadata_v0["tables"][child_table_name]["fields"][child_foreign_key][
            "subtype"
        ] = "string"
    return metadata_v0



import numpy as np
import pandas as pd
from sdmetrics.goal import Goal
from sdmetrics.utils import is_datetime

from syntherela.metrics.base import SingleColumnMetric, DistanceBaseMetric
from syntherela.metrics.single_column.distance.utils import get_histograms


class TotalVariationDistance(DistanceBaseMetric, SingleColumnMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "TotalVariationDistance"
        self.goal = Goal.MINIMIZE
        self.min_value = 0.0
        self.max_value = float("inf")

    @staticmethod
    def is_applicable(column_type):
        return column_type in ["categorical", "numerical", "datetime", "boolean"]

    @staticmethod
    def compute(real_data, synthetic_data, bins, **kwargs):
        """
        Total Variation Distance metric.
        """

        f_exp, f_obs = get_histograms(
            real_data, synthetic_data, normalize=True, bins=bins
        )
        total_variation = 0
        for i in range(len(f_obs)):
            total_variation += abs(f_obs[i] - f_exp[i])

        return total_variation

    def run(self, real_data, synthetic_data, **kwargs):
        if self.is_constant(real_data):
            return {
                "value": 0,
                "reference_ci": [0, 0],
                "bootstrap_mean": 0,
                "bootstrap_se": 0,
            }
        # check for datetime
        if is_datetime(real_data):
            real_data = pd.to_numeric(real_data, errors="coerce", downcast="integer")
            synthetic_data = pd.to_numeric(
                synthetic_data, errors="coerce", downcast="integer"
            )
        if real_data.dtype.name in ("object", "category", "bool"):
            bins = None
        else:
            real_data = real_data.dropna()
            synthetic_data = synthetic_data.dropna()
            bins = np.histogram_bin_edges(real_data)
        return super().run(real_data, synthetic_data, bins=bins, **kwargs)



from syntherela.metadata import drop_ids
from syntherela.metrics.base import DetectionBaseMetric, SingleTableMetric


class SingleTableDetection(DetectionBaseMetric, SingleTableMetric):
    def prepare_data(self, real_data, synthetic_data, metadata, **kwargs):
        real_data = drop_ids(real_data, metadata)
        synthetic_data = drop_ids(synthetic_data, metadata)
        return super().prepare_data(real_data, synthetic_data)



"""Metadata Exceptions."""


class MetadataError(Exception):
    """Error to raise when Metadata is not valid."""


class MetadataNotFittedError(MetadataError):
    """Error to raise when Metadata is used before fitting."""




from relsyndgb.benchmark import Benchmark
import argparse
import logging
import sys

from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from xgboost import XGBClassifier

from relsyndgb.metrics.single_column.distance import HellingerDistance, JensenShannonDistance, WassersteinDistance, TotalVariationDistance
from relsyndgb.metrics.single_column.statistical import ChiSquareTest, KolmogorovSmirnovTest
from relsyndgb.metrics.single_table.distance import MaximumMeanDiscrepancy, PairwiseCorrelationDifference
from relsyndgb.metrics.single_column.detection import SingleColumnDetection
from relsyndgb.metrics.single_table.detection import SingleTableDetection
from relsyndgb.metrics.multi_table.detection import DenormalizedDetection, DenormalizedAggregationDetection, SingleTableAggregationDetection, ParentChildDetection, ParentChildAggregationDetection
from relsyndgb.metrics.multi_table.statistical import CardinalityShapeSimilarity

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="airbnb-simplified_subsampled")
args.add_argument("--methods",'-m', action='append', default=None)
args = args.parse_args()
dataset_name = args.dataset_name
methods = args.methods

logger = logging.getLogger(f'{dataset_name}_logger')

logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.info(f"START LOGGING Dataset: {dataset_name}")

xgb_cls = XGBClassifier
xgb_args = {'seed': 0}
rf_cls = RandomForestClassifier
rf_args = {'random_state': 0, 'n_estimators': 100}
logistic = LogisticRegression
logistic_args = {'random_state': 0}
single_column_metrics = [
    # ChiSquareTest(), 
    # KolmogorovSmirnovTest(),
    # TotalVariationDistance(),
    # HellingerDistance(),
    # JensenShannonDistance(),
    # WassersteinDistance(),
    # SingleColumnDetection(classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42),
    # # SingleColumnDetection(classifier_cls=rf_cls, classifier_args=rf_args),
    # SingleColumnDetection(classifier_cls=logistic, classifier_args=logistic_args, random_state=42),
    ]
single_table_metrics = [
    # MaximumMeanDiscrepancy(),
    # PairwiseCorrelationDifference(),
    # SingleTableDetection(classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42),
    # # SingleTableDetection(classifier_cls=rf_cls, classifier_args=rf_args),
    # SingleTableDetection(classifier_cls=logistic, classifier_args=logistic_args, random_state=42),
                        ]
multi_table_metrics = [
    # CardinalityShapeSimilarity(),
    # # DenormalizedDetection(classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42),
    # # DenormalizedDetection(classifier_cls=rf_cls, classifier_args=rf_args),
    # # DenormalizedDetection(classifier_cls=logistic, classifier_args=logistic_args, random_state=42),
    # # DenormalizedAggregationDetection(classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42),
    # # DenormalizedAggregationDetection(classifier_cls=rf_cls, classifier_args=rf_args),
    # # DenormalizedAggregationDetection(classifier_cls=logistic, classifier_args=logistic_args, random_state=42),
    # SingleTableAggregationDetection(classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42),
    # # SingleTableAggregationDetection(classifier_cls=rf_cls, classifier_args=rf_args),
    # SingleTableAggregationDetection(classifier_cls=logistic, classifier_args=logistic_args, random_state=42),
    # ParentChildDetection(classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42),
    # ParentChildDetection(classifier_cls=logistic, classifier_args=logistic_args, random_state=42),
    ParentChildAggregationDetection(classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42),
    ParentChildAggregationDetection(classifier_cls=logistic, classifier_args=logistic_args, random_state=42),
    ]

benchmark = Benchmark(
    real_data_dir='/Users/martinjurkovic/Documents/github_projects/relsyndgb/data/downloads',
    synthetic_data_dir='/Users/martinjurkovic/Documents/github_projects/relsyndgb/data/synthetic',
    results_dir='/Users/martinjurkovic/Documents/github_projects/relsyndgb/experiments/evaluation/results',
    benchmark_name='Evaluation_1',
    single_column_metrics=single_column_metrics,
    single_table_metrics=single_table_metrics,
    multi_table_metrics=multi_table_metrics,
    run_id="1",
    sample_id="sample1",
    # datasets=[dataset_name],
    datasets = ['CORA_v1']
    # methods=methods
)

benchmark.run()
# benchmark.read_results()

# benchmark.reports['Biodegradability_v1']['MOSTLYAI'].visualize_distributions()

# benchmark.visualize_single_column_metrics()
# benchmark.visualize_single_table_metrics()
# benchmark.visualize_multi_table_metrics()
